from __future__ import absolute_import
from pyspark.sql import SparkSession
from py4j.protocol import Py4JJavaError

from .extractor import Extractor
from .tools import remove_hdfs_prefix, fix_suffix, infer_input_path_from_partition


class JSONExtractor(Extractor):
    """
    The JSONExtractor class provides an API to extract data stored as JSON format,
    deserializes it into a PySpark dataframe and returns it. Currently only
    single-line JSON files are supported, stored either as textFile or sequenceFile.

    Examples
    --------
    >>> from spooq import extractor as E

    >>> extractor = E.JSONExtractor(input_path="tests/data/schema_v1/sequenceFiles")
    >>> extractor.input_path == "tests/data/schema_v1/sequenceFiles" + "/*"
    True

    >>> extractor = E.JSONExtractor(
    >>>     base_path="tests/data/schema_v1/sequenceFiles",
    >>>     partition="20200201"
    >>> )
    >>> extractor.input_path == "tests/data/schema_v1/sequenceFiles" + "/20/02/01" + "/*"
    True

    Parameters
    ----------
    input_path : :any:`str`
        The path from which the JSON files should be loaded ("/\\*" will be added if omitted)
    base_path : :any:`str`
        Spooq tries to infer the ``input_path`` from the ``base_path`` and the ``partition`` if the
        ``input_path`` is missing.
    partition : :any:`str` or :any:`int`
        Spooq tries to infer the ``input_path`` from the ``base_path`` and the ``partition`` if the
        ``input_path`` is missing.
        Only daily partitions in the form of "YYYYMMDD" are supported. e.g., "20200201" => <base_path> + "/20/02/01/\\*"

    Returns
    -------
    |SPARK_DATAFRAME|
        The extracted data set as a PySpark DataFrame

    Raises
    ------
    :any:`AttributeError`
        Please define either ``input_path`` or ``base_path`` and ``partition``

    Warning
    ---------
    Currently only single-line JSON files stored as SequenceFiles or TextFiles are supported!

    Note
    ----
    The init method checks which input parameters are provided and derives the final input_path
    from them accordingly.

    If ``input_path`` is not :any:`None`:
        Cleans ``input_path`` and returns it as the final ``input_path``

    Elif ``base_path`` and ``partition`` are not :any:`None`:
        Cleans ``base_path``, infers the sub path from the ``partition``
        and returns the combined string as the final ``input_path``

    Else:
        Raises an :any:`AttributeError`

    """

    def __init__(self, input_path=None, base_path=None, partition=None):
        super(JSONExtractor, self).__init__()
        self.input_path = self._get_path(input_path=input_path, base_path=base_path, partition=partition)
        self.base_path = base_path
        self.partition = partition
        self.spark = (
            SparkSession.Builder()
            .enableHiveSupport()
            .appName("spooq.extractor: {nm}".format(nm=self.name))
            .getOrCreate()
        )

    def extract(self):
        """
        This is the Public API Method to be called for all classes of Extractors

        Parameters
        ----------

        Returns
        -------
        |SPARK_DATAFRAME|
            Complex PySpark DataFrame deserialized from the input JSON Files

        """
        self.logger.info("Loading Raw RDD from: " + self.input_path)
        rdd_raw = self._get_raw_rdd(self.input_path)
        rdd_strings = self._get_values_as_string_rdd(rdd_raw)
        return self._convert_rdd_to_df(rdd_strings)

    def _get_path(self, input_path=None, base_path=None, partition=None):
        """
        Checks which input parameters are provided and derives the final input_path from them.

        If :py:data:`input_path` is not :any:`None`:
            Cleans :py:data:`input_path` and returns it as the final input_path

        If :py:data:`base_path` and :py:data:`partition` are not :any:`None`:
            Cleans :py:data:`base_path`, infers the sub path from the :py:data:`partition` and returns
            the combined String as the final input_path

        If none of the above holds true, an Exception is raised

        Parameters
        ----------
        input_path : :any:`str`
        base_path : :any:`str`
        partition : :any:`str` or :any:`int`

        Returns
        -------
        :any:`str`
            The final input_path to be used for Extraction.

        Raises
        ------
        :py:class:`AttributeError`
            Please define either (input_path) or (base_path and partition)

        Examples
        --------
        >>> _get_path(input_path=u'/user/furia_salamandra_faerfax/data')
        u'/user/furia_salamandra_faerfax/data/*'

        >>> _get_path(base_path=u'/user/furia_salamandra_faerfax/data', partition=20180101)
        u'/user/furia_salamandra_faerfax/data/18/01/01/*'

        See Also
        --------
        :py:meth:`fix_suffix`
        :py:meth:`remove_hdfs_prefix`

        """
        if input_path:
            return fix_suffix(remove_hdfs_prefix(input_path))
        elif base_path and partition:
            return infer_input_path_from_partition(base_path=base_path, partition=partition)
        else:
            error_msg = "Please define either (input_path) or (base_path and partition)"
            self.logger.error(error_msg)
            raise AttributeError(error_msg)

    def _get_raw_rdd(self, input_path):
        """
        Loads TextFiles containing JSON Strings from :py:data:`input_path` and parallelizes them
        into an :any:pyspark.RDD`.

        Parameters
        ----------
        input_path : :any:`str`

        Returns
        -------
        :any:`pyspark.RDD`
            Output RDD with one JSON String per Record

        See Also
        --------
        pyspark.SparkContext.textFile

        """
        try:
            return self._get_raw_sequence_rdd(input_path)
        except Py4JJavaError:
            return self._get_raw_text_rdd(input_path)

    def _get_raw_text_rdd(self, input_path):
        self.logger.debug("Fetching TextFile containing JSON")
        return self.spark.sparkContext.textFile(input_path)

    def _get_raw_sequence_rdd(self, input_path):
        self.logger.debug("Fetching SequenceFile containing JSON")
        return self.spark.sparkContext.sequenceFile(input_path).map(lambda k_v: k_v[1].decode("utf-8"))

    def _convert_rdd_to_df(self, rdd_strings):
        """
        Converts the input RDD :py:data:`rdd_strings` to a DataFrame with inferred Structure
        and DataTypes and returns it.

        Parameters
        ----------
        rdd_strings : :any:`pyspark.RDD`
            Input RDD containing only unicode JSON Strings per Record

        Returns
        -------
        |SPARK_DATAFRAME|
            Complex DataFrame with set of all found Attributes from input JSON Files

        See Also
        --------
        pyspark.sql.DataFrameReader.json

        """
        self.logger.debug("Deserializing JSON from String RDD to DataFrame")
        return self.spark.read.json(rdd_strings)

    def _get_values_as_string_rdd(self, rdd_raw):
        """
        Removes newline ``\n`` and carriage return ``\r`` characters.

        Parameters
        ----------
        rdd_raw : :any:`pyspark.RDD`

        Returns
        -------
        :any:`pyspark.RDD`
            Output RDD with one JSON String per Record

        """
        self.logger.debug("Cleaning JSON String RDD (selecting values, " + "removing newline and carriage return)")
        return rdd_raw.map(lambda v: v.replace("\\n", " ").replace("\\r", ""))



"""
Global Logger instance used by spooq.

Example
-------
>>> import logging
>>> logga = logging.getLogger("spooq")
<logging.Logger at 0x7f5dc8eb2890>
>>> logga.info("Hello World")
[spooq] 2020-03-21 23:55:48,253 INFO logging_example::<module>::4: Hello World

"""

import os
import sys
import logging

from spooq._version import __version__ as version_number

initialized = False


def initialize():
    """
    Initializes the global logger for Spooq with pre-defined levels for ``stdout`` and ``stderr``.
    No input parameters are needed, as the configuration is received via :py:meth:`get_logging_level`.

    Note
    ----
    The output format is defined as:
        | "[%(name)s] %(asctime)s %(levelname)s %(module)s::%(funcName)s::%(lineno)d: %(message)s"
        | For example "[spooq] 2020-03-11 15:40:59,313 DEBUG newest_by_group::__init__::53: group by columns: [u'user_id']"

    Warning
    -------
    The ``root`` logger of python is also affected as it has to have a level at least as
    fine grained as the logger of Spooq, to be able to produce an output.
    """
    global initialized
    if initialized:
        return

    logging_level = get_logging_level()

    # logging.getLogger("root").setLevel(logging_level)
    logger = logging.getLogger("spooq")
    logger.setLevel(logging_level)

    if not len(logger.handlers):
        formatter = logging.Formatter(
            "[%(name)s] %(asctime)s %(levelname)s %(module)s::%(funcName)s::%(lineno)d: %(message)s"
        )

        # STDOUT Handler
        ch_out = logging.StreamHandler(sys.stdout)
        ch_out.setLevel(logging_level)
        ch_out.setFormatter(formatter)
        logger.addHandler(ch_out)

        # STDERR Handler
        # ch_err = logging.StreamHandler(sys.stderr)
        # ch_err.setLevel(logging_level)
        # ch_err.setFormatter(formatter)
        # logger.addHandler(ch_err)

        initialized = True

    logger.info(f"Thank you for choosing Spooq {version_number}!")


def get_logging_level():
    """
    Returns the logging level depending on the environment variable `SPOOQ_ENV`.

    Note
    ----
    If SPOOQ_ENV is
        * **dev**        -> "DEBUG"
        * **test**       -> "ERROR"
        * something else -> "INFO"

    Returns
    -------
    :any:`str`
        Logging level
    """
    spooq_env = os.getenv("SPOOQ_ENV", "default").lower()
    if spooq_env.startswith("dev"):
        return "DEBUG"
    elif spooq_env.startswith("test"):
        return "ERROR"
    elif spooq_env.startswith("pr"):
        return "WARN"
    else:
        return "INFO"



import logging
from pathlib import Path
from typing import Dict, Union

from pyspark.sql import functions as F, types as T, DataFrame, SparkSession

from spooq.shared import EnumMode
from spooq.transformer.transformer import Transformer


class AnnotatorMode(EnumMode):
    """Possible values: ['insert', 'upsert']"""

    insert = "Only add missing comments. Don't overwrite existing comments"
    upsert = "Insert missing comments. Overwrite existing comments"


class MissingColumnHandling(EnumMode):
    """Possible values: ['raise_error', 'skip']"""

    raise_error = "Raise an error when a comment was defined for a missing column"
    skip = "Skip / ignore each comment that refers to a missing column"


class ColumnNotFound(ValueError):
    pass


def load_comments_from_metastore_table(sql_source_table_identifier: str) -> Dict[str, str]:
    """
    Extracts comments from metadata stored in table (loaded by path or metastore).

    Parameters
    ----------
    sql_source_table_identifier : :any:`str`
        This is the fully qualified table name used for the ``DESCRIBE <sql_source_table_identifier>`` SQL command.
        Some examples:

            * delta.`/path/to/table/files.delta`
            * my_db.my_table
            * my_catalog.my_db.my_table

    Returns
    -------
    :any:`dict`
        Dictionary containing all extracted, non-null comments per column. ({<column_name>: <comment>})
    """
    spark = SparkSession.Builder().getOrCreate()
    table_description = spark.sql(f"DESCRIBE {sql_source_table_identifier}")
    return {k: v for k, v in table_description.rdd.map(lambda row: (row.col_name, row.comment)).collect() if v}


def update_comment(
    df: DataFrame,
    column_name: str,
    comment: str,
    annotator_mode: AnnotatorMode = AnnotatorMode.upsert,
    missing_column_handling: MissingColumnHandling = MissingColumnHandling.raise_error,
    logger: logging.Logger = None,
) -> DataFrame:
    """
    Updates a single column's comment within the provided dataframe.

    Parameters
    ----------
        df : |SPARK_DATAFRAME|
            The dataframe that contains the column to be updated.
        column_name : :any:`str`
            The name of the column to be commented.
        comment : :any:`str`
            The comment to apply to the referenced column.
        annotator_mode : :py:class:`AnnotatorMode`, defaults to AnnotatorMode.upsert
            Defines if existing columns get overwritten.
        missing_column_handling : :py:class:`MissingColumnHandling`, defaults to ``MissingColumnHandling.raise_error``
            Defines how to behave when the ``column_name`` is not found in the dataframe.

    Raises:
        ColumnNotFound: If column is missing from the dataframe and missing_column_handling is set to ``raise_error``

    Returns:
        |SPARK_DATAFRAME|: Dataframe with updated metadata (comment) for the specified column.
    """
    logger = logger or logging.getLogger("spooq")
    if column_name not in df.columns:
        if missing_column_handling == MissingColumnHandling.skip:
            logger.warning(
                f"Column '{column_name}' was not found in the dataframe. MissingColumnHandling is set to 'skip'"
            )
            return df
        else:
            raise ColumnNotFound(
                f"Comment <{comment}> cannot be applied to Column <{column_name}> "
                "because the column was not found in the dataframe:\n"
                "\n".join(df.columns)
            )

    existing_comment = df.schema[column_name].metadata.get("comment")
    if annotator_mode == AnnotatorMode.insert and existing_comment:
        logger.info(
            f"A comment already exists for column '{column_name}' but the annotation mode is set to insert only "
            f"-> skipping the update. Existing comment: '{existing_comment}', New comment: '{comment}'"
        )
        return df

    return df.withColumn(column_name, F.col(column_name).alias(column_name, metadata={"comment": comment}))


class Annotator(Transformer):
    """
    Inserts or upserts column comments to dataframes. Can also be used just to fetch column comments
    from an existing table.

    Parameters
    ----------
    comments_mapping  : :class:`dict`, Defaults to {}
        Dictionary consisting of column names and comments
    mode : :py:class:`AnnotatorMode`, Defaults to ``AnnotatorMode.upsert``
        This mode defines how the transformer should react to existing column comments. ``insert`` will leave existing
        untouched while ``upsert`` overwrites them if a new comment is provided. Existing columns are defined as
        comments already defined in the dataframe or the sql_source_table!
    missing_column_handling : :py:class:`MissingColumnHandling`, Defaults to ``MissingColumnHandling.raise_error``
        This mode defines how the transformer should react to missing columns that are referenced in
        the ``comments_mapping``.
    sql_source_table_identifier : :any:`str`, Defaults to None
        The transformer will load any existing comments from the defined source table to the ``comments_mapping``

    Example
    -------
    >>> '''Fetching comments from an existing source table and applying those plus explicitely defined comments'''
    >>> # Schema of sql_source_table (/tmp/path/to/my/silver_table.delta):
    >>> #     col_A string COMMENT "Comment from sql_source_table",
    >>> #     col_B string COMMENT "Comment from sql_source_table",
    >>> #     col_Z string COMMENT "Comment from sql_source_table",
    >>>
    >>> # Schema of input dataframe
    >>> #     col_A string,
    >>> #     col_B string,
    >>> #     col_Y string,
    >>>
    >>> from spooq.transformer import Annotator
    >>> from spooq.transformer.annotator import AnnotatorMode, MissingColumnHandling
    >>>
    >>> spark.createDataFrame(
    >>>     [],
    >>>     schema='''
    >>>         col_A string COMMENT "Comment from sql_source_table",
    >>>         col_B string COMMENT "Comment from sql_source_table",
    >>>         col_Z string COMMENT "Comment from sql_source_table"
    >>>     '''
    >>> ).write.format("delta").mode("overwrite").options(mergeSchema=True).save("/tmp/path/to/my/silver_table.delta")
    >>> input_df = spark.createDataFrame([], "col_A string, col_B string, col_Y string")
    >>> comments_mapping = {
    >>>     "col_A": "Updated comment from comments_mapping",
    >>>     "col_Y": "New comment from comments_mapping",
    >>> }
    >>>
    >>> output_df = Annotator(
    >>>     comments_mapping=comments_mapping,
    >>>     mode=AnnotatorMode.upsert,
    >>>     missing_column_handling=MissingColumnHandling.raise_error,
    >>>     sql_source_table_identifier="delta.`/tmp/path/to/my/silver_table.delta`",
    >>> ).transform(input_df)
    >>>
    >>> print(json.dumps({col["name"]: col["metadata"]["comment"] for col in output_df.schema.J["fields"]}, indent=2))
    {
    "col_A": "Updated comment from comments_mapping",
    "col_B": "Comment from sql_source_table",
    "col_Y": "New comment from comments_mapping"
    }

    Note
    ----
    Here are some use cases to use this transformer:
        - Add explicitely defined comments from ``comments_mapping`` to dataframe within silver pipeline
        - Apply comments from existing silver table to dataframe within gold pipeline
        - Apply comments from ``columns_mapping`` within the :py:class:`~spooq.transformer.mapper.Mapper` transformer
        - Apply comments from existing silver table to gold dataframe
          within the :py:class:`~spooq.transformer.mapper.Mapper` transformer
        - Combine any of them

    Raises
    ------
        ColumnNotFound: If column is missing from the dataframe and missing_column_handling is set to ``raise_error``
    """

    def __init__(
        self,
        comments_mapping: Dict[str, str] = None,
        mode: AnnotatorMode = AnnotatorMode.upsert,
        missing_column_handling: MissingColumnHandling = MissingColumnHandling.raise_error,
        sql_source_table_identifier: str = None,
    ):
        super().__init__()
        self.comments_mapping = comments_mapping or {}
        self.mode = mode
        self.missing_column_handling = missing_column_handling
        self.sql_source_table_identifier = sql_source_table_identifier

    def transform(self, input_df: DataFrame) -> DataFrame:
        if self.sql_source_table_identifier:
            self.logger.info(
                f"sql_source_table_identifier was provided: "
                f"Trying to load existing comments from source table at <{self.sql_source_table_identifier}> ..."
            )
            existing_comments_mapping = load_comments_from_metastore_table(
                sql_source_table_identifier=self.sql_source_table_identifier
            )
            self.logger.info(
                f"Trying to add {len(existing_comments_mapping)} existing comments "
                f"from {self.sql_source_table_identifier} to provided comments_mapping!"
            )
            for column, existing_comment in existing_comments_mapping.items():
                if not column in input_df.columns:
                    self.logger.debug(
                        f"Skipping to apply comment ('{existing_comment}') fetched from the sql_source_table because "
                        f"column: '{column}' was not found in the dataframe."
                    )
                    continue

                if column in self.comments_mapping:
                    self.logger.debug(
                        f"Comment for column: '{column}' is both defined in the sql_source_table ('{existing_comment}') "
                        f"and the mapping ('{self.comments_mapping[column]}')."
                    )
                    if self.mode == AnnotatorMode.upsert:
                        self.logger.debug("Using comment from mapping.")
                        continue

                self.logger.debug("Using comment from sql_source_table.")
                self.comments_mapping[column] = existing_comment

            self.logger.info(
                f"Added {len(set(input_df.columns).intersection(set(existing_comments_mapping)))} "
                f"comments from {self.sql_source_table_identifier}!"
            )

        self.logger.info(f"Applying {len(self.comments_mapping)} comments from mapping to DataFrame ...")
        for column, comment in self.comments_mapping.items():
            input_df = update_comment(
                df=input_df,
                column_name=column,
                comment=comment,
                annotator_mode=self.mode,
                missing_column_handling=self.missing_column_handling,
                logger=self.logger,
            )
        return input_df



"""
This abstract class provides the functionality to log any cleansed values into a separate column
that contains a struct with a sub column per cleansed column (according to the `cleaning_definition`).
If a value was cleansed, the original value will be stored in its respective sub column. If a value was not
cleansed, the sub column will be empty (None).
"""

from pyspark.sql import functions as F, types as T

from .transformer import Transformer


class BaseCleaner(Transformer):
    def __init__(
        self,
        cleaning_definitions,
        column_to_log_cleansed_values,
        store_as_map=False,
        temporary_columns_prefix="1b75cdd2e2356a35486230c69cfac5493488a919",
    ):
        super().__init__()
        self.cleaning_definitions = cleaning_definitions
        self.column_to_log_cleansed_values = column_to_log_cleansed_values
        self.store_as_map = store_as_map
        self.TEMPORARY_COLUMNS_PREFIX = temporary_columns_prefix

    def _get_temporary_column_names(self, column_names):
        return [f"{self.TEMPORARY_COLUMNS_PREFIX}_{column_name}" for column_name in column_names]

    @staticmethod
    def _add_temporary_columns(input_df, column_names, temporary_column_names):
        for column_name, temporary_column_name in zip(column_names, temporary_column_names):
            # copy columns to be cleansed to temporary column
            input_df = input_df.withColumn(temporary_column_name, F.col(column_name))
        return input_df

    def _log_cleansed_values(self, input_df, column_names, temporary_column_names):
        def _only_keep_cleansed_values(col_name, temporary_col_name):
            return (F.when(F.col(temporary_col_name) == F.col(col_name), F.lit(None))
                .otherwise(F.when(F.col(temporary_col_name).isNull() & F.col(col_name).isNull(), F.lit(None))
                .otherwise(F.when(F.col(temporary_col_name).isNull() & F.col(col_name).isNotNull(), F.lit("null"))
                .otherwise(F.col(temporary_col_name).try_cast("string"))))
            )

        for column_name, temporary_column_name in zip(column_names, temporary_column_names):
            # Only keep cleansed values in temporary columns
            input_df = input_df.withColumn(
                temporary_column_name, _only_keep_cleansed_values(column_name, temporary_column_name)
            )

        input_df = input_df.withColumn(
            self.column_to_log_cleansed_values,
            F.struct(
                [
                    F.col(temp_col_name).alias(col_name)
                    for col_name, temp_col_name in zip(column_names, temporary_column_names)
                ]
            ),
        )

        if self.store_as_map:
            input_df = input_df.withColumn(
                self.column_to_log_cleansed_values, F.to_json(self.column_to_log_cleansed_values)
            )
            input_df = input_df.withColumn(
                self.column_to_log_cleansed_values,
                F.from_json(self.column_to_log_cleansed_values, T.MapType(T.StringType(), T.StringType())),
            )
            # set empty map to null
            input_df = input_df.withColumn(
                self.column_to_log_cleansed_values,
                F.when(F.size(F.col(self.column_to_log_cleansed_values))>0, F.col(self.column_to_log_cleansed_values)).otherwise(F.lit(None)),
            )

        return input_df.drop(*temporary_column_names)



from __future__ import absolute_import
from builtins import str
from pyspark.sql.window import Window  # noqa: F401
from pyspark.sql.functions import row_number, when

from .transformer import Transformer


class NewestByGroup(Transformer):
    """
    Groups, orders and selects first element per group.

    Example
    -------
    >>> transformer = NewestByGroup(
    >>>     group_by=["first_name", "last_name"],
    >>>     order_by=["created_at_ms", "version"]
    >>> )

    Parameters
    ----------
    group_by : :any:`str` or :any:`list` of :any:`str`, (Defaults to ['id'])
        List of attributes to be used within the Window Function as Grouping Arguments.

    order_by : :any:`str` or :any:`list` of :any:`str`, (Defaults to ['updated_at', 'deleted_at'])
        List of attributes to be used within the Window Function as Ordering Arguments.
        All columns will be sorted in **descending** order.

    Raises
    ------
    :any:`exceptions.AttributeError`
        If any Attribute in :py:data:`group_by` or :py:data:`order_by` is not contained in the
        input DataFrame.

    Note
    ----
    PySpark's :py:class:`~pyspark.sql.Window` function is used internally
    The first row (:py:meth:`~pyspark.sql.functions.row_number`) per window will be selected and returned.
    """

    def __init__(self, group_by=["id"], order_by=["updated_at", "deleted_at"]):
        super(NewestByGroup, self).__init__()
        self.group_by = []
        if isinstance(group_by, list):
            self.group_by.extend(group_by)
        else:
            self.group_by.extend([group_by])

        self.order_by = []
        if isinstance(order_by, list):
            self.order_by.extend(order_by)
        else:
            self.order_by.extend([order_by])

        self.logger.debug("group by columns: " + str(self.group_by))
        self.logger.debug("order by columns: " + str(self.order_by))

    def transform(self, input_df):
        self.logger.debug(
            """{grp_by} used for grouping, {ord_by} used for ordering""".format(
                grp_by=self.group_by, ord_by=self.order_by
            )
        )

        window = self._construct_window_function(input_df, self.group_by, self.order_by)

        return input_df.select("*", row_number().over(window).alias("row_nr")).where("row_nr = 1").drop("row_nr")

    def _construct_window_function(self, input_df, group_by, order_by):
        """Constructs a window function based on the given input params"""
        group_by_query = [input_df[col] for col in group_by]
        order_by_query = [input_df[col].desc_nulls_last() for col in order_by]

        return Window.partitionBy(group_by_query).orderBy(order_by_query)



"""
This type of object glues the aforementioned processes together and  extracts, transforms
(Transformer chain possible) and loads the data from start to end.
"""

from builtins import str
from builtins import object
import logging


class Pipeline(object):
    """
    Represents a Pipeline of an Extractor, (multiple) Transformers and a Loader Object.

    Attributes
    ----------
        extractor : Subclass of :py:class:`spooq.extractor.Extractor`
            The entry point of the Pipeline. Extracts a DataFrame from a Source.
        transformers : List of Subclasses of :py:class:`spooq.transformer.Transformer` Objects
            The Data Wrangling Part of the Pipeline. A chain of Transformers, a single Transformer
            or a PassThrough Transformer can be set and used.
        loader : Subclass of :py:class:`spooq.loader.Loader`
            The exit point of the Pipeline. Loads a DataFrame to a target Sink.
        name : :any:`str`
            Sets the `__name__` of the class' type as `name`, which is essentially the Class' Name.
        logger : :any:`logging.Logger`
            Shared, class level logger for all instances.

    Example
    -------
    >>> from spooq.pipeline import Pipeline
    >>> import spooq.extractor as   E
    >>> import spooq.transformer as T
    >>> import spooq.loader as      L
    >>>
    >>> #  Definition how the output table should look like and where the attributes come from:
    >>> users_mapping = [
    >>>     ("id",              "id",                     "IntegerType"),
    >>>     ("guid",            "guid",                   "StringType"),
    >>>     ("forename",        "attributes.first_name",  "StringType"),
    >>>     ("surename",        "attributes.last_name",   "StringType"),
    >>>     ("gender",          "attributes.gender",      "StringType"),
    >>>     ("has_email",       "attributes.email",       "StringBoolean"),
    >>>     ("has_university",  "attributes.university",  "StringBoolean"),
    >>>     ("created_at",      "meta.created_at_ms",     "timestamp_ms_to_s"),
    >>> ]
    >>>
    >>> #  The main object where all steps are defined:
    >>> users_pipeline = Pipeline()
    >>>
    >>> #  Defining the EXTRACTION:
    >>> users_pipeline.set_extractor(E.JSONExtractor(
    >>>     input_path="tests/data/schema_v1/sequenceFiles"
    >>> ))
    >>>
    >>> #  Defining the TRANSFORMATION:
    >>> users_pipeline.add_transformers([
    >>>     T.Mapper(mapping=users_mapping),
    >>>     T.ThresholdCleaner(thresholds={"created_at": {
    >>>                                                 "min": 0,
    >>>                                                 "max": 1580737513,
    >>>                                                 "default": None}}),
    >>>     T.NewestByGroup(group_by="id", order_by="created_at")
    >>> ])
    >>>
    >>> #  Defining the LOAD:
    >>> users_pipeline.set_loader(L.HiveLoader(
    >>>     db_name="users_and_friends",
    >>>     table_name="users",
    >>>     partition_definitions=[{
    >>>         "column_name": "dt",
    >>>         "column_type": "IntegerType",
    >>>         "default_value": 20200201}],
    >>>     repartition_size=10,
    >>> ))
    >>>
    >>> #  Executing the whole ETL pipeline
    >>> users_pipeline.execute()
    """

    def __init__(self, input_df=None, bypass_loader=False):
        if input_df is not None:
            self.bypass_extractor = True
            self.input_df = input_df
        else:
            self.bypass_extractor = False

        self.bypass_loader = bypass_loader
        self.extractor = None
        self.transformers = []
        self.loader = None

        self.name = type(self).__name__
        self.logger = logging.getLogger("spooq")
        self.logger.info("New {cls_name} Instance created\n".format(cls_name=str(self.name)) + str(self))

    def execute(self):
        """
        Executes the whole Pipeline at once.

        Extracts from the Source, transformes the DataFrame and loads it into a target Sink.

        Returns
        -------
        input_df : |SPARK_DATAFRAME|
            **If** the ``bypass_loader`` attribute was set to True in the Pipeline class,
            the output DataFrame from the Transformer(s) will be directly returned.

        Note
        ----
        This method does not take ANY input parameters. All needed parameters are defined
        at the initialization phase.
        """
        extracted_df = self.extract()
        transformed_df = self.transform(extracted_df)
        return self.load(transformed_df)

    def extract(self):
        """Calls the extract Method on the Extractor Object.

        Returns
        -------
        |SPARK_DATAFRAME|
            The output_df from the Extractor used as the input for the Transformer (chain).
        """
        if self.bypass_extractor:
            return self.input_df
        else:
            return self.extractor.extract()

    def transform(self, input_df):
        """
        Calls the transform Method on the Transformer Object(s) in the order of importing the
        Objects while passing the DataFrame.

        Parameters
        ----------
        input_df : |SPARK_DATAFRAME|
            The output DataFrame of the Extractor Object.

        Returns
        -------
        |SPARK_DATAFRAME|
            The input DataFrame for the Loader.
        """
        transformed_df = input_df

        for transformer in self.transformers:
            transformed_df = transformer.transform(transformed_df)

        return transformed_df

    def load(self, input_df):
        """
        Calls the load Method on the Loader Object.

        Parameters
        ----------
        input_df : |SPARK_DATAFRAME|
            The output DataFrame from the Transformer(s).

        Returns
        -------
        input_df : |SPARK_DATAFRAME|
            **If** the ``bypass_loader`` attribute was set to True in the Pipeline class,
            the output DataFrame from the Transformer(s) will be directly returned.
        """
        if self.bypass_loader:
            return input_df
        else:
            return self.loader.load(input_df)

    def set_extractor(self, extractor):
        """
        Sets an Extractor Object to be used within the Pipeline.

        Parameters
        ----------
        extractor : Subclass of :py:class:`spooq.extractor.Extractor`
            An already initialized Object of any Subclass of spooq.extractor.Extractor.

        Raises
        ------
        :any:`exceptions.AssertionError`:
            An input_df was already provided which bypasses the extraction action
        """
        assert not self.bypass_extractor, "An input_df was already provided which bypasses the extraction action"
        self.extractor = extractor

    def add_transformers(self, transformers):
        """
        Adds a list of Transformer Objects to be used within the Pipeline.

        Parameters
        ----------
        transformer : :any:`list` of Subclass of :py:class:`spooq.transformer.Transformer`
            Already initialized Object of any Subclass of spooq.transformer.Transformer.
        """

        try:
            self.transformers.extend(transformers)
        except TypeError:
            self.transformers.append(transformers)

    def clear_transformers(self):
        """
        Clears the list of already added Transformers.
        """
        self.transformers = []

    def set_loader(self, loader):
        """
        Sets an Loader Object to be used within the Pipeline.

        Parameters
        ----------
        loader : Subclass of :py:class:`spooq.loader.Loader`
            An already initialized Object of any Subclass of spooq.loader.Loader.

        Raises
        ------
        :any:`exceptions.AssertionError`:
            You can not set a loader if the `bypass_loader` parameter is set.
        """
        assert not self.bypass_loader, "You can not set a loader if the `bypass_loader` parameter is set."
        self.loader = loader

    def __str__(self):
        return """spooq.pipeline Object
                Used Extractor:    {extr}
                Used Transformers: {trans}
                Used Loader:       {ldr}
                """.format(
            extr=str(self.extractor),
            trans=str(["{e}\n".format(e=extr) for extr in self.transformers]),
            ldr=str(self.loader),
        )



from spooq.transformer.newest_by_group import NewestByGroup
from spooq.transformer.mapper import Mapper
from spooq.transformer.exploder import Exploder
from spooq.transformer.threshold_cleaner import ThresholdCleaner
from spooq.transformer.enum_cleaner import EnumCleaner
from spooq.transformer.null_cleaner import NullCleaner
from spooq.transformer.sieve import Sieve
from spooq.transformer.annotator import Annotator

__all__ = [
    "NewestByGroup",
    "Mapper",
    "Exploder",
    "ThresholdCleaner",
    "EnumCleaner",
    "Sieve",
    "NullCleaner",
    "Annotator",
]



from __future__ import absolute_import
from past.builtins import basestring
from pyspark.sql import SparkSession
from pyspark.sql import types as sql_types
from pyspark.sql.functions import lit

from .loader import Loader


class HiveLoader(Loader):
    """
    Persists a PySpark DataFrame into a Hive Table.

    Examples
    --------
    >>> HiveLoader(
    >>>     db_name="users_and_friends",
    >>>     table_name="friends_partitioned",
    >>>     partition_definitions=[{
    >>>         "column_name": "dt",
    >>>         "column_type": "IntegerType",
    >>>         "default_value": 20200201}],
    >>>     clear_partition=True,
    >>>     repartition_size=10,
    >>>     overwrite_partition_value=False,
    >>>     auto_create_table=False,
    >>> ).load(input_df)

    >>> HiveLoader(
    >>>     db_name="users_and_friends",
    >>>     table_name="all_friends",
    >>>     partition_definitions=[],
    >>>     repartition_size=200,
    >>>     auto_create_table=True,
    >>> ).load(input_df)

    Parameters
    ----------
    db_name : :any:`str`
        The database name to load the data into.
    table_name : :any:`str`
        The table name to load the data into. The database name must not be included in this
        parameter as it is already defined in the `db_name` parameter.
    partition_definitions : :any:`list` of :py:class:`dict`
        (Defaults to `[{"column_name": "dt", "column_type": "IntegerType", "default_value": None}]`).

            * **column_name** (:any:`str`) - The Column's Name to partition by.
            * **column_type** (:any:`str`) - The PySpark SQL DataType for the Partition Value as
              a String. This should normally either be 'IntegerType()' or 'StringType()'
            * **default_value** (:any:`str` or :any:`int`) - If `column_name` does not contain
              a value or `overwrite_partition_value` is set, this value will be used for the
              partitioning

    clear_partition : :any:`bool`, (Defaults to True)
        This flag tells the Loader to delete the defined partitions before
        inserting the input DataFrame into the target table. Has no effect if no partitions are
        defined.
    repartition_size : :any:`int`, (Defaults to 40)
        The DataFrame will be repartitioned on Spark level before inserting into the table.
        This effects the number of output files on which the Hive table is based.
    auto_create_table : :any:`bool`, (Defaults to True)
        Whether the target table will be created if it does not yet exist.
    overwrite_partition_value : :any:`bool`, (Defaults to True)
        Defines whether the values of columns defined in `partition_definitions` should
        explicitly set by default_values.

    Raises
    ------
    :any:`exceptions.AssertionError`:
        partition_definitions has to be a list containing dicts. Expected dict content:
        'column_name', 'column_type', 'default_value' per partition_definitions item.

    :any:`exceptions.AssertionError`:
        Items of partition_definitions have to be dictionaries.

    :any:`exceptions.AssertionError`:
        No column name set!

    :any:`exceptions.AssertionError`:
        Not a valid (PySpark) datatype for the partition column {name} | {type}.

    :any:`exceptions.AssertionError`:
        `clear_partition` is only supported if `overwrite_partition_value` is also enabled.
        This would otherwise result in clearing partitions on basis of dynamically values
        (from DataFrame) instead of explicitly defining the partition(s) to clear.
    """

    def __init__(
        self,
        db_name,
        table_name,
        partition_definitions=[{"column_name": "dt", "column_type": "IntegerType", "default_value": None}],
        clear_partition=True,
        repartition_size=40,
        auto_create_table=True,
        overwrite_partition_value=True,
    ):
        super(HiveLoader, self).__init__()
        self._assert_partition_definitions_is_valid(partition_definitions)
        self.partition_definitions = partition_definitions
        self.db_name = db_name
        self.table_name = table_name
        self.full_table_name = db_name + "." + table_name
        self.repartition_size = repartition_size
        if clear_partition and not overwrite_partition_value:
            raise ValueError(
                "clear_partition is only supported if overwrite_partition_value is also enabled. ",
                "This would otherwise result in clearing partitions on basis of dynamically values",
                "(from dataframe) instead of explicitly defining the partition(s) to clear",
            )
        self.clear_partition = clear_partition
        self.overwrite_partition_value = overwrite_partition_value
        self.auto_create_table = auto_create_table
        self.spark = (
            SparkSession.Builder()
            .enableHiveSupport()
            .appName("spooq.extractor: {nm}".format(nm=self.name))
            .getOrCreate()
        )

    def _assert_partition_definitions_is_valid(self, definitions):
        assert isinstance(definitions, list), (
            "partition_definitions has to be a list containing dicts.\n",
            "Expected dict content: 'column_name', 'column_type', 'default_value' per partition_definitions item.",
        )

        for dct in definitions:
            assert isinstance(dct, dict), "Items of partition_definitions have to be dictionaries"
            assert dct["column_name"], "No column name set!"
            assert isinstance(dct["column_type"], basestring) and hasattr(
                sql_types, dct["column_type"]
            ), "Not a valid (PySpark) datatype for the partition column {name} | {type}".format(
                name=dct["column_name"], type=dct["column_type"]
            )
            dct["column_type"] = getattr(sql_types, dct["column_type"])

    def load(self, input_df):
        self.spark.conf.set("hive.exec.dynamic.partition", "true")
        self.spark.conf.set("hive.exec.dynamic.partition.mode", "nonstrict")

        input_df = input_df.repartition(self.repartition_size)
        input_df = self._add_partition_definition_to_dataframe(input_df)

        if self._table_exists():
            output_df = self.spark.table(self.full_table_name)
            assert input_df.columns == output_df.columns, "Input columns don't match the columns of the Hive table"
            if self.clear_partition:
                self._clear_hive_partition()
            input_df.write.insertInto(self.full_table_name)
        elif self.auto_create_table:
            input_df.write.partitionBy(*[dct["column_name"] for dct in self.partition_definitions]).saveAsTable(
                self.full_table_name
            )

        else:
            raise Exception(
                "Table: {tbl} does not exist and `auto_create_table` is set to False".format(tbl=self.full_table_name)
            )

    def _add_partition_definition_to_dataframe(self, input_df):
        for partition_definition in self.partition_definitions:
            if partition_definition["column_name"] not in input_df.columns or self.overwrite_partition_value:
                assert "default_value" in list(partition_definition.keys()) and (
                    partition_definition["default_value"] or partition_definition["default_value"] == 0
                ), "No default partition value set for partition column: {name}!\n".format(
                    name=partition_definition["column_name"]
                )
                input_df = input_df.withColumn(
                    partition_definition["column_name"],
                    lit(partition_definition["default_value"]).try_cast(partition_definition["column_type"]()),
                )
        return input_df

    def _table_exists(self):
        table_exists = False
        self.spark.catalog.setCurrentDatabase(self.db_name)
        for tbl in self.spark.catalog.listTables():
            if self.table_name == tbl.name:
                table_exists = True
        return table_exists

    def _clear_hive_partition(self):
        def _construct_partition_query_string(partition_definitions):
            partition_queries = []
            for dct in partition_definitions:
                assert "default_value" in list(
                    dct.keys()
                ), "clear_partitions needs a default_value per partition definition!"
                if issubclass(dct["column_type"], sql_types.NumericType):
                    partition_queries.append("{part} = {dt}".format(part=dct["column_name"], dt=dct["default_value"]))
                else:
                    partition_queries.append("{part} = '{dt}'".format(part=dct["column_name"], dt=dct["default_value"]))
            return ", ".join(partition_queries)

        partition_query = _construct_partition_query_string(self.partition_definitions)
        command = """ALTER TABLE {tbl} DROP IF EXISTS PARTITION ({part_def})""".format(
            tbl=self.full_table_name, part_def=partition_query
        )

        self.logger.debug("Command used to clear Partition: {cmd}".format(cmd=command))
        self.spark.sql(command)



from __future__ import absolute_import
from past.builtins import basestring
from .transformer import Transformer


class Sieve(Transformer):
    """
    Filters rows depending on provided filter expression.
    Only records complying with filter condition are kept.

    Examples
    --------
    >>> transformer = T.Sieve(filter_expression=\"\"\" attributes.last_name rlike "^.{7}$" \"\"\")

    >>> transformer = T.Sieve(filter_expression=\"\"\" lower(gender) = "f" \"\"\")

    Parameters
    ----------
    filter_expression  : :any:`str`
        A valid PySpark SQL expression which returns a boolean

    Raises
    ------
    :any:`exceptions.ValueError`
         filter_expression has to be a valid (Spark)SQL expression provided as a string

    Note
    ----
    The :meth:`~pyspark.sql.DataFrame.filter` method is used internally.

    Note
    ----
    The Size of the resulting DataFrame is not guaranteed to be equal to the Input DataFrame!
    """

    def __init__(self, filter_expression):
        super(Sieve, self).__init__()
        self.filter_expression = filter_expression

        if not isinstance(filter_expression, basestring):
            raise ValueError(
                "filter_expression has to be a valid (Spark)SQL expression ",
                "provided as a string",
            )

    def transform(self, input_df):
        return input_df.filter(self.filter_expression)



from functools import partial
from types import FunctionType
from typing import Union, List, Tuple, Any, Callable
import warnings

from pyspark.errors.exceptions.base import AnalysisException, ParseException
from pyspark.sql import (
    functions as F,
    types as T,
    DataFrame,
    Column,
)

from spooq.shared import EnumMode
from spooq.transformer.annotator import Annotator
from spooq.transformer.transformer import Transformer
from spooq.transformer.mapper_transformations import as_is


class MapperMode(EnumMode):
    """Possible values: ['replace', 'append', 'prepend', 'rename_and_validate']"""

    replace = "output schema = columns from mapping"
    append = "output schema = input columns + columns from mapping"
    prepend = "output schema = columns from mapping + input columns"
    rename_and_validate = "output schema = columns from mapping + validation"


class MissingColumnHandling(EnumMode):
    """Possible values: ['raise_error', 'skip', 'nullify']"""

    raise_error = "Raise an exception"
    skip = "Skip the mapping transformation"
    nullify = "Create source column filled with null"


class MapperParameterNotSupported(AttributeError):
    pass


class DataTypeValidationFailed(ValueError):
    pass


class ColumnMappingNotSupported(ValueError):
    pass


class Mapper(Transformer):
    """
    Selects, transforms, comments and casts or validates a DataFrame based on the provided mapping.

    Examples
    --------
    >>> from pyspark.sql import functions as F, types as T
    >>> from spooq.transformer import Mapper
    >>> from spooq.transformer import mapper_transformations as spq
    >>>
    >>> cmt_id = "Identifier for entity (UUID4)"
    >>> cmt_type = "Type of entity ('type_a', 'type_b' or 'type_c')"
    >>>
    >>> mapping = [
    >>>     ("id",            "data.relationships.food.data.id",  spq.to_str,                     cmt_id),
    >>>     ("version",       "data.version",                     spq.to_int),
    >>>     ("type",          "elem.attributes.type",             "string",                       cmt_type),
    >>>     ("created_at",    "elem.attributes.created_at",       spq.to_timestamp),
    >>>     ("created_on",    "elem.attributes.created_at",       spq.to_timestamp(cast="date")),
    >>>     ("processed_at",  F.current_timestamp(),              spq.as_is,
    >>> ]
    >>> mapper = Mapper(mapping=mapping)
    >>> mapper.transform(input_df).printSchema()
    root
     |-- id: string (nullable = true)
     |-- version: integer (nullable = true)
     |-- type: string (nullable = true)
     |-- created_at: timestamp (nullable = true)
     |-- created_on: date (nullable = true)
     |-- processed_at: timestamp (nullable = false)

    Parameters
    ----------
    mapping  : :class:`list` of :any:`tuple` containing three or four elements, respectively
        This is the main parameter for this transformation. It gives information
        about the column names for the output DataFrame, the column names (paths)
        from the input DataFrame, their data types and optionally a column comment.
        Custom data types are also supported, which can clean, pivot, anonymize, ...
        the data itself. Please have a look at the
        :py:mod:`spooq.transformer.mapper_transformations` module for more information.

    missing_column_handling : :py:class:`MissingColumnHandling`, Defaults to ``MissingColumnHandling.raise_error``
        Specifies how to proceed in case a source column does not exist in the source DataFrame:
            * raise_error (default)
                Raise an exception
            * nullify
                Create source column filled with null
            * skip
                Skip the mapping transformation

    ignore_ambiguous_columns : :any:`bool`, Defaults to False
        It can happen that the input DataFrame has ambiguous column names (like "Key" vs "key") which will
        raise an exception with Spark when reading. This flag surpresses this exception and skips those affected
        columns.

    mode : :py:class:`MapperMode`, Defaults to ``MapperMode.replace``
        Defines whether the mapping should fully replace the schema of the input DataFrame or just add to it.
        Following modes are supported:

            * replace
                The output schema is the same as the provided mapping.
                => output schema: columns from mapping
            * append
                The columns provided in the mapping are added at the end of the input schema. If a column already
                exists in the input DataFrame, its position is kept.
                => output schema: input columns + columns from mapping
            * prepend
                The columns provided in the mapping are added at the beginning of the input schema. If a column already
                exists in the input DataFrame, its position is kept.
                => output schema: columns from mapping + input columns
            * rename_and_validate
                All built-in, custom transformations (except renaming) and casts are disabled. The Mapper only
                renames the columns and validates that the output data type is the same as the input data type. The
                transformation will fail if any spooq / custom transformations (except `as_is`) are defined!
                => output schema: columns from mapping

    annotator_option : :any:`dict`, Defaults to {}
        Options that are passed as parameters to the Annotator instance used by the Mapper
        Transformer if comments are provided.

    Keyword Arguments
    -----------------

    Note
    ----
    Let's talk about Mappings:

    The mapping should be a list of tuples that contain all necessary information per column.

    * Column Name: :any:`str`
        Sets the name of the column in the resulting output DataFrame.
    * Source Path / Name / Column / Function: :any:`str`, :class:`~pyspark.sql.Column` or :mod:`~pyspark.sql.functions`
        Points to the name of the column in the input DataFrame. If the input
        is a flat DataFrame, it will essentially be the column name. If it is of complex
        type, it will point to the path of the actual value. For example: ``data.relationships.sample.data.id``,
        where id is the value we want. It is also possible to directly pass
        a PySpark Column which will get evaluated. This can contain arbitrary logic supported by Spark. For example:
        ``F.current_date()`` or ``F.when(F.col("size") == 180, F.lit("tall")).otherwise(F.lit("tiny"))``.
    * DataType: :any:`str`, :class:`~pyspark.sql.types.DataType` or :mod:`~spooq.transformer.mapper_transformations`
        DataTypes can be types from :py:mod:`pyspark.sql.types` (like T.StringType()),
        simple strings supported by PySpark (like "string") and custom transformations provided by spooq
        (like spq.to_timestamp). You can find more information about the transformations at
        https://spooq.rtfd.io/en/latest/transformer/mapper.html#module-spooq.transformer.mapper_transformations.
    * Comment: :any:`str` (optional)
        Applies a comment to the respective column, if provided.

    Note
    ----
    The available input columns can vary from batch to batch if you use schema inference
    (f.e. on json data) for the extraction. Via the parameter `missing_column_handling` you can specify a strategy on
    how to handle missing columns on the input DataFrame.
    It is advised to use the 'raise_error' option as it can uncover typos and bugs.
    """

    def __init__(
        self,
        mapping: List[Tuple],
        ignore_ambiguous_columns: bool = False,
        missing_column_handling: MissingColumnHandling = MissingColumnHandling.raise_error,
        mode: MapperMode = MapperMode.replace,
        annotator_options: dict|None = None,
        **kwargs,
    ):
        super(Mapper, self).__init__()
        self.mapping = mapping
        self.missing_column_handling = missing_column_handling
        self.ignore_ambiguous_columns = ignore_ambiguous_columns
        self.mode = mode
        self.annotator_options = annotator_options or {}

        if isinstance(self.mode, str):
            message = f"Mapper parameter: `mode` as string is not supported anymore, please provide the mode as a MapperMode object!"
            raise MapperParameterNotSupported(message)

        if isinstance(self.missing_column_handling, str):
            message = (
                "Mapper parameter: `missing_column_handling` as string is not supported anymore, "
                "please provide the parameter as a MissingColumnHandling object!"
            )
            raise MapperParameterNotSupported(message)

        if "ignore_missing_columns" in kwargs:
            message = (
                "Mapper parameter: `ignore_missing_columns` is not supported anymore, "
                "use `missing_column_handling=MissingColumnHandling.nullify` instead!")
            raise MapperParameterNotSupported(message)

    def transform(self, input_df: DataFrame) -> DataFrame:
        self.logger.info("Generating SQL Select-Expression for Mapping...")
        self.logger.debug("Input Schema/Mapping:")
        self.logger.debug("\n" + "\n".join(["\t".join(map(str, mapping_line)) for mapping_line in self.mapping]))

        if not self.mapping:
            raise ColumnMappingNotSupported("It seems like the provided mapping is empty!")

        input_columns = input_df.columns
        select_expressions = []
        with_column_expressions = []
        validation_errors = []
        comments_mapping = {}

        for column_mapping in self.mapping:
            if not len(column_mapping) in [3, 4]:
                raise ColumnMappingNotSupported(
                    "Each mapping tuple (column) must contain at least 3 and at most 4 elements!\n"
                    f"Following mapping contains {len(column_mapping)} elements: {column_mapping}"
                )

            name: str = column_mapping[0]
            source_column_raw: str | Column = column_mapping[1]
            data_transformation_raw: str | Callable = column_mapping[2]
            comment: str | None = None
            if len(column_mapping) == 4:
                comment: str | None = column_mapping[3]

            self.logger.debug(f"generating Select statement for attribute: {name}")
            self.logger.debug(
                f"generate mapping for name: {name}, "
                f"source_column: {source_column_raw}, "
                f"data_transformation: {data_transformation_raw}, "
                f"comment: {comment}"
            )

            source_column: Column | None = self._get_source_spark_column(source_column_raw, name, input_df)
            if source_column is None:
                if self.mode == MapperMode.rename_and_validate:
                    validation_errors.append(DataTypeValidationFailed(f"Input column: {name} not found!"))
                continue

            source_spark_data_type: T.DataType | None = self._get_spark_data_type(input_df.select(source_column).dtypes[0][1])
            target_transformation: T.DataType | Column | None = self._get_spark_data_type(
                data_transformation_raw
            ) or self._get_spooq_transformation(name, source_column, data_transformation_raw)

            if self.mode == MapperMode.rename_and_validate:
                try:
                    self._validate_data_type(
                        name,
                        source_spark_data_type,
                        data_transformation_raw,
                        target_transformation,
                    )
                except DataTypeValidationFailed as e:
                    self.logger.warning(e)
                    validation_errors.append(e)

            if isinstance(target_transformation, T.DataType):
                if target_transformation != source_spark_data_type:
                    select_expression = source_column.try_cast(target_transformation).alias(name)
                else:
                    select_expression = source_column.alias(name)
            else:
                select_expression = target_transformation

            self.logger.debug(f"Select-Expression for Attribute {name}: {select_expression}")
            if self.mode.name in ["append", "prepend"] and name in input_columns:
                with_column_expressions.append((name, select_expression))
            else:
                select_expressions.append(select_expression)

            if comment:
                comments_mapping[name] = comment

        if validation_errors:
            raise DataTypeValidationFailed(f"Validation has failed!\n{validation_errors}")

        self.logger.info("SQL Select-Expression for new mapping generated!")
        self.logger.debug("SQL Select-Expressions for new mapping:\n" + "\n".join(str(select_expressions).split(",")))
        self.logger.debug("SQL WithColumn-Expressions for new mapping: " + str(with_column_expressions))

        if self.mode == MapperMode.prepend:
            df_to_return = input_df.select(select_expressions + ["*"])
        elif self.mode == MapperMode.append:
            df_to_return = input_df.select(["*"] + select_expressions)
        else:
            df_to_return = input_df.select(select_expressions)

        if with_column_expressions:
            for name, expression in with_column_expressions:
                df_to_return = df_to_return.withColumn(name, expression)

        if comments_mapping or self.annotator_options.get("sql_source_table_identifier"):
            annotator = Annotator(comments_mapping=comments_mapping, **self.annotator_options)
            df_to_return = annotator.transform(df_to_return)

        return df_to_return

    def _get_source_spark_column(
        self, source_column: Union[str, Column], name: str, input_df: DataFrame
    ) -> Union[Column, None]:
        """
        Returns the provided source column as a Column.
        Supports source column definition as a string or a Column (including functions).
        """
        try:
            input_df.select(source_column)
            if isinstance(source_column, str):
                source_column = F.col(source_column)

        except AnalysisException as e:
            if isinstance(source_column, str) and self.missing_column_handling == MissingColumnHandling.skip:
                self.logger.warning(
                    f"Missing column ({str(source_column)}) skipped (MissingColumnHandling.skip): {e.getMessage()}"
                )
                return None
            elif isinstance(source_column, str) and self.missing_column_handling == MissingColumnHandling.nullify:
                self.logger.warning(
                    f"Missing column ({str(source_column)}) replaced with NULL (via MissingColumnHandling.nullify): {e.getMessage()}"
                )
                source_column = F.lit(None)
            elif "ambiguous" in e.getMessage().lower() and self.ignore_ambiguous_columns:
                self.logger.warning(
                    f'Exception ignored (via ignore_ambiguous_columns=True) for column "{source_column}": {e.getMessage()}'
                )
                return None
            else:
                self.logger.exception(
                    f"Column: '{source_column}' cannot be resolved but is referenced in the mapping by "
                    f"column: '{name}'. You can make use of the following parameters to handle "
                    f"missing input columns:\n{MissingColumnHandling.to_string()}"
                )
                raise e
        return source_column

    def _get_spark_data_type(self, data_transformation: Union[str, Column, Callable, T.DataType]) -> T.DataType | None:
        """
        Returns the PySpark data type as Python object (None if not found). Supports Python objects and strings.
        """
        data_type = None

        if isinstance(data_transformation, T.DataType):
            data_type = data_transformation  # Spark datatype as Python object

        elif isinstance(data_transformation, str):
            data_type_ = data_transformation.replace("()", "")
            try:
                data_type = getattr(T, data_type_)()  # Spark datatype as string
            except AttributeError:
                try:
                    data_type = T._parse_datatype_string(
                        "void" if data_type_ == "null" else data_type_
                    )  # Spark datatype as short string
                except ParseException:
                    pass

        return data_type

    def _validate_data_type(
        self,
        name: str,
        source_spark_data_type: T.DataType | None,
        original_data_transformation: str | T.DataType | Callable,
        target_transformation: T.DataType | Column,
    ):
        """
        Validates that the input and output data types are the same. NULL sources and as_is transformations are valid.
        """
        if isinstance(target_transformation, T.DataType):
            if isinstance(source_spark_data_type, T.NullType):
                self.logger.warning(
                    f"The data type of column {name} could not be validated because the source data type is NULL!"
                )
            elif target_transformation != source_spark_data_type:
                raise DataTypeValidationFailed(
                    f"The mode is set to `rename_and_validate` but the target data_type ({target_transformation}) "
                    f"does not match the source data_type ({source_spark_data_type}) for the column: {name}!"
                )
            else:
                self.logger.debug(f"No validation errors found for column: {name}")
        else:
            if original_data_transformation == as_is:
                pass
            else:
                raise DataTypeValidationFailed(
                    f"Spooq transformations are not allowed in 'rename_and_validate' mode! Name: {name}, transformation: {original_data_transformation}"
                )

    def _get_spooq_transformation(
        self, name: str, source_column: str | Column, data_transformation: str | Callable
    ) -> Column:
        """
        Applies the defined spooq transformation and returns the target column.
        """
        self.logger.debug(
            f"Get spooq transformation for column: {name} (source: {source_column}) with "
            f"following requested transformation: {data_transformation}"
        )
        if isinstance(data_transformation, FunctionType):
            # function without brackets / parameters (f.e.: `spq.as_is`)
            spooq_partial = data_transformation()
            return spooq_partial(source_column=source_column, name=name)

        elif isinstance(data_transformation, partial):
            # function with brackets / parameters (f.e.: `spq.as_is()` or `spq.as_is(cast="string")`)
            args = data_transformation.keywords
            args.setdefault("source_column", source_column)
            args.setdefault("name", name)
            return data_transformation(**args)
        else:
            raise ColumnMappingNotSupported(
                f"The provided data transformation for column: {name} is not supported: {data_transformation}!"
            )



import pyspark.sql.functions as F
from pyspark.sql.column import Column

from .base_cleaner import BaseCleaner


class NullCleaner(BaseCleaner):
    """
    Fills Null values of the specifield fields.
    Takes a dictionary with the fields to be cleaned and the default value to be set when the field is null.

    Examples
    --------
    >>> from pyspark.sql import functions as F
    >>> from spooq.transformer import NullCleaner
    >>> transformer = NullCleaner(
    >>>     cleaning_definitions={
    >>>         "points": {
    >>>             "default": 0
    >>>         }
    >>>     }
    >>> )

    >>> from spooq.transformer import NullCleaner
    >>> from pyspark.sql import Row
    >>>
    >>> input_df = spark.createDataFrame([
    >>>     Row(id=0, points=5),
    >>>     Row(id=1, points= None),
    >>>     Row(id=2, points=15),
    >>> ])
    >>> transformer = NullCleaner(
    >>>     cleaning_definitions={
    >>>         "points":    {"default":  0},
    >>>     },
    >>>     column_to_log_cleansed_values="cleansed_values_null",
    >>>     store_as_map=True,
    >>> )
    >>> output_df = transformer.transform(input_df)
    >>> output_df.show()
    +---+------+--------------------+
    | id|points|cleansed_values_null|
    +---+------+--------------------+
    |  0|     5|                null|
    |  1|     0|    [points -> null]|
    |  2|    15|                null|
    +---+------+--------------------+
    >>> output_df.printSchema()
     |-- id: long (nullable = true)
     |-- points: long (nullable = true)
     |-- cleansed_values_null: map (nullable = true)
     |    |-- key: string
     |    |-- value: string (valueContainsNull = true)

    Parameters
    ----------
    cleaning_definitions : :py:class:`dict`
        Dictionary containing column names and respective default values

    column_to_log_cleansed_values : :any:`str`, Defaults to None
        Defines a column in which the original (uncleansed) value will be stored in case of cleansing. If no column
        name is given, nothing will be logged.

    store_as_map : :any:`bool`, Defaults to False
        Specifies if the logged cleansed values should be stored in a column as :any:`pyspark.sql.types.MapType` or as
        :any:`pyspark.sql.types.StructType` with stringified values.

    Note
    ----
    The following cleaning_definitions attributes per column are mandatory:

        * default - :class:`~pyspark.sql.column.Column` or any primitive Python value
            If a value gets cleansed it gets replaced with the provided default value.

    Returns
    -------
    |SPARK_DATAFRAME|
        The transformed DataFrame

    Raises
    ------
    :any:`exceptions.ValueError`
        Null-based cleaning requires the field default.
        Default parameter is not specified for column with name: {column_name}
    """

    def __init__(self, cleaning_definitions=None, column_to_log_cleansed_values=None, store_as_map=False):
        cleaning_definitions = cleaning_definitions if cleaning_definitions is not None else {}
        super().__init__(
            cleaning_definitions=cleaning_definitions,
            column_to_log_cleansed_values=column_to_log_cleansed_values,
            store_as_map=store_as_map,
            temporary_columns_prefix="ead8cn9f7tf0sf1cs1ua61464zti82kj",
        )
        self.logger.debug("Cleansing Definitions: " + str(self.cleaning_definitions))

    def transform(self, input_df):
        self.logger.debug("input_df Schema: " + input_df._jdf.schema().treeString())
        ordered_column_names = input_df.columns

        if self.column_to_log_cleansed_values:
            column_names_to_clean = self.cleaning_definitions.keys()
            temporary_column_names = self._get_temporary_column_names(column_names_to_clean)
            input_df = self._add_temporary_columns(input_df, column_names_to_clean, temporary_column_names)

        cleansing_expressions = []
        for column_name, cleansing_value in list(self.cleaning_definitions.items()):

            data_type = input_df.schema[str(column_name)].dataType
            substitute = cleansing_value.get("default")

            if substitute is None:
                raise ValueError("Null-based cleaning requires the field default.", f"Default parameter is not specified for column with name: {column_name}")

            if not isinstance(substitute, Column):
                substitute = F.lit(substitute)

            self.logger.debug("Cleansing value for column " + column_name + ": " + str(cleansing_value))

            cleansing_expression = (
                F.when(
                    F.col(column_name).isNull(),
                    substitute,
                ).otherwise(F.col(column_name))
                .try_cast(data_type)
            )
            self.logger.debug("Cleansing Expression for " + column_name + ": " + str(cleansing_expression))
            cleansing_expressions.append((column_name, cleansing_expression))

        self.logger.info("Full null cleansing expression:")
        self.logger.info(
            ".".join(
                [
                    f"withColumn({column_name}, {str(cleansing_expr)})"
                    for (column_name, cleansing_expr) in cleansing_expressions
                ]
            )
        )
        for (column_name, cleansing_expr) in cleansing_expressions:
            input_df = input_df.withColumn(column_name, cleansing_expr)

        if self.column_to_log_cleansed_values:
            input_df = self._log_cleansed_values(input_df, column_names_to_clean, temporary_column_names)
            ordered_column_names.append(self.column_to_log_cleansed_values)

        return input_df.select(ordered_column_names)



from __future__ import absolute_import
import sys

if sys.version_info.major > 2:
    # This is needed for python 2 as otherwise pyspark raises an exception for following command:
    # data_type = input_df.schema[str(column_name)].dataType
    # Pyspark checks if the input is a string, which does not work
    # with the new strings from builtins
    from builtins import str
import pyspark.sql.functions as F
import pyspark.sql.types as sql_types
from pyspark.sql.column import Column

from .base_cleaner import BaseCleaner


class ThresholdCleaner(BaseCleaner):
    """
    Cleanes values based on defined boundaries and optionally logs the original values.
    The valid value ranges are provided via dictionary for each column to be cleaned. Boundaries can either be of
    static or dynamic nature that resolve to numbers, timetamps or dates.

    Examples
    --------
    >>> from pyspark.sql import functions as F
    >>> from spooq.transformer import ThresholdCleaner
    >>> transformer = ThresholdCleaner(
    >>>     thresholds={
    >>>         "created_at": {
    >>>             "min": F.date_sub(F.current_date(), 365 * 10),  # Ten years ago
    >>>             "max": F.current_date()
    >>>         },
    >>>         "size_cm": {
    >>>             "min": 70,
    >>>             "max": 250,
    >>>             "default": None
    >>>         },
    >>>     }
    >>> )

    >>> from spooq.transformer import ThresholdCleaner
    >>> from pyspark.sql import Row
    >>>
    >>> input_df = spark.createDataFrame([
    >>>     Row(id=0, integers=-5, doubles=-0.75),
    >>>     Row(id=1, integers= 5, doubles= 1.25),
    >>>     Row(id=2, integers=15, doubles= 0.67),
    >>> ])
    >>> transformer = ThresholdCleaner(
    >>>     thresholds={
    >>>         "integers": {"min":  0, "max": 10},
    >>>         "doubles":  {"min": -1, "max":  1}
    >>>     },
    >>>     column_to_log_cleansed_values="cleansed_values_threshold",
    >>>     store_as_map=True,
    >>> )
    >>> output_df = transformer.transform(input_df)
    >>> output_df.show(truncate=False)
    +---+--------+-----------+---------------------+
    |id |integers|doubles|cleansed_values_threshold|
    +---+--------+-------+-------------------------+
    |0  |null    |-0.75  |{integers -> -5}         |
    |1  |5       |null   |{doubles -> 1.25}        |
    |2  |null    |0.67   |{integers -> 15}         |
    +---+--------+-------+-------------------------+
    >>> output_df.printSchema()
     |-- id: long (nullable = true)
     |-- integers: long (nullable = true)
     |-- doubles: double (nullable = true)
     |-- cleansed_values_threshold: map (nullable = false)
     |    |-- key: string
     |    |-- value: string (valueContainsNull = true)

    Parameters
    ----------
    thresholds : :py:class:`dict`
        Dictionary containing column names and respective valid ranges

    column_to_log_cleansed_values : :any:`str`, Defaults to None
        Defines a column in which the original (uncleansed) value will be stored in case of cleansing. If no column
        name is given, nothing will be logged.

    store_as_map : :any:`bool`, Defaults to False
        Specifies if the logged cleansed values should be stored in a column as :class:`~pyspark.sql.types.MapType` with
        stringified values or as :class:`~pyspark.sql.types.StructType` with the original respective data types.

    Note
    ----
    Following cleansing rule attributes per column are supported:

        * min, mandatory: :any:`str`, :any:`int`, |SPARK_COLUMN|, |SPARK_FUNCTION|
            A number or timestamp/date which serves as the lower limit for allowed values. Values
            below this threshold will be cleansed. Supports literals, Spark functions and Columns.
        * max, mandatory: :any:`str`, :any:`int`, |SPARK_COLUMN|, |SPARK_FUNCTION|
            A number or timestamp/date which serves as the upper limit for allowed values. Values
            above this threshold will be cleansed. Supports literals, Spark functions and Columns.
        * default, defaults to None: :any:`str`, :any:`int`, |SPARK_COLUMN|, |SPARK_FUNCTION|
            If a value gets cleansed it gets replaced with the provided default value. Supports literals,
            Spark functions and Columns.

    The :py:meth:`~pyspark.sql.Column.between` method is used internally.

    Returns
    -------
    |SPARK_DATAFRAME|
        The transformed DataFrame

    Raises
    ------
    :any:`exceptions.ValueError`
        Threshold-based cleaning only supports Numeric, Date and Timestamp Types!
        Column with name: {col_name} and type of: {col_type} was provided

    Warning
    -------
    Only Numeric, TimestampType, and DateType data types are supported!
    """

    def __init__(self, thresholds={}, column_to_log_cleansed_values=None, store_as_map=False):
        super().__init__(
            cleaning_definitions=thresholds,
            column_to_log_cleansed_values=column_to_log_cleansed_values,
            store_as_map=store_as_map,
            temporary_columns_prefix="dac28b56d8055953a7038bfe3b5097e7",
        )
        self.logger.debug("Range Definitions: " + str(self.cleaning_definitions))

    def transform(self, input_df):
        self.logger.debug("input_df Schema: " + input_df._jdf.schema().treeString())
        ordered_column_names = input_df.columns

        if self.column_to_log_cleansed_values:
            column_names_to_clean = self.cleaning_definitions.keys()
            temporary_column_names = self._get_temporary_column_names(column_names_to_clean)
            input_df = self._add_temporary_columns(input_df, column_names_to_clean, temporary_column_names)

        cleansing_expressions = []
        for column_name, value_range in list(self.cleaning_definitions.items()):

            data_type = input_df.schema[str(column_name)].dataType
            substitute = value_range.get("default", None)
            if not isinstance(substitute, Column):
                substitute = F.lit(substitute)

            if not isinstance(data_type, (sql_types.NumericType, sql_types.DateType, sql_types.TimestampType)):
                raise ValueError(
                    "Threshold-based cleaning only supports Numeric, Date and Timestamp Types!\n",
                    "Column with name: {col_name} and type of: {col_type} was provided".format(
                        col_name=column_name, col_type=data_type
                    ),
                )

            self.logger.debug("Ranges for column " + column_name + ": " + str(value_range))

            cleansing_expression = (
                F.when(
                    input_df[column_name].between(value_range["min"], value_range["max"]),
                    input_df[column_name],
                )
                .otherwise(substitute)
                .try_cast(data_type)
            )
            self.logger.debug("Cleansing Expression for " + column_name + ": " + str(cleansing_expression))
            cleansing_expressions.append((column_name, cleansing_expression))

        self.logger.info("Full threshold cleansing expression:")
        self.logger.info(
            ".".join(
                [
                    f"withColumn({column_name}, {str(cleansing_expr)})"
                    for (column_name, cleansing_expr) in cleansing_expressions
                ]
            )
        )
        for (column_name, cleansing_expr) in cleansing_expressions:
            input_df = input_df.withColumn(column_name, cleansing_expr)

        if self.column_to_log_cleansed_values:
            input_df = self._log_cleansed_values(input_df, column_names_to_clean, temporary_column_names)
            ordered_column_names.append(self.column_to_log_cleansed_values)

        return input_df.select(ordered_column_names)



"""
Extractors are used to fetch, extract and convert a source data set into a PySpark DataFrame.
Exemplary extraction sources are **JSON Files** on file systems like HDFS, DBFS or EXT4
and relational database systems via **JDBC**.
"""

from builtins import object
import logging


class Extractor(object):
    """
    Base Class of Extractor Classes.

    Attributes
    ----------
    name : :any:`str`
        Sets the `__name__` of the class' type as `name`, which is essentially the Class' Name.
    logger : :any:`logging.Logger`
        Shared, class level logger for all instances.

    """

    def __init__(self):
        self.name = type(self).__name__
        self.logger = logging.getLogger("spooq")

    def extract(self):
        """
        Extracts Data from a Source and converts it into a PySpark DataFrame.

        Returns
        -------
        |SPARK_DATAFRAME|

        Note
        ----
        This method does not take ANY input parameters. All needed parameters are defined
        in the initialization of the Extractor Object.
        """
        raise NotImplementedError("This method has to be implemented in the subclasses")

    def __str__(self):
        return "Extractor Object of Class {nm}".format(nm=self.name)



"""
This is a collection of module level functions to be applied to a DataFrame.
These methods can be used with the :py:class:`~spooq.transformer.mapper.Mapper` transformer
or directly within a ``select`` or ``withColumn`` statement.

All functions support following generic functionalities:

    - ``alt_src_cols``: Alternative source columns that will be used within a coalesce function if provided
    - ``cast``: Explicit casting after the transformation (sane defaults are set for each function)

``to_str``, ``to_int``, ``to_long``, ``to_float``, ``to_double`` are convenience methods with
a hardcoded cast that cannot be changed.

All examples assume following code has been executed before:

>>> from pyspark.sql import Row
>>> from pyspark.sql import functions as F, types as T
>>> from spooq.transformer import Mapper
>>> from spooq.transformer import mapper_transformations as spq
"""
import re
from functools import partial
from typing import Any, Union, List, Tuple, Set, Callable
import json

from pyspark.sql import functions as F, types as T
from pyspark.sql.column import Column


COLUMN_NAME_PATTERN = re.compile(r".*\'(.*)\'")


def _coalesce_source_columns(
    source_column: Union[str, Column],
    alt_src_cols: Union[str, Column, List[Union[str, Column]], Tuple[Union[str, Column]], Set[Union[str, Column]]],
) -> Column:
    if alt_src_cols is None or alt_src_cols is False:
        return source_column
    if not isinstance(alt_src_cols, (list, tuple, set)):
        return F.coalesce(source_column, alt_src_cols)
    return F.coalesce(source_column, *alt_src_cols)


def _get_executable_function(
    inner_func: Callable, source_column: Union[str, Column], name: str, **kwargs
) -> Union[partial, Column]:
    direct_call = True

    if isinstance(source_column, str):
        name = name or source_column.split(".")[-1]
        source_column = F.col(source_column)
    elif isinstance(source_column, Column):
        name = name or re.search(COLUMN_NAME_PATTERN, source_column.name()).group(1)
    else:
        direct_call = False

    if direct_call:
        return inner_func(source_column, name, **kwargs)
    else:
        return partial(inner_func, **kwargs)


def _datatype_is_of_integer_family(datatype: T.DataType | str) -> bool:
    if isinstance(datatype, T.DataType):
        datatype = datatype.simpleString()

    return datatype in ["tinyint", "smallint", "int", "bigint"]

def as_is(source_column: Union[str, Column] = None, name: str = None, **kwargs) -> Union[partial, Column]:
    """
    Returns a renamed column without any casting. This is especially useful if you need to
    keep a complex data type (f.e. array, list or struct).

    Parameters
    ----------
    source_column : str or Column
        Input column. Can be a name, pyspark column or pyspark function
    name : str, default -> derived from input column
        Name of the output column. (``.alias(name)``)

    Keyword Arguments
    -----------------
    alt_src_cols : str, default -> no coalescing, only source_column
        Coalesce with source_column and columns from this parameter.
    cast : T.DataType(), default -> no casting, same return data type as input data type
        Applies provided datatype on output column (``.try_cast(cast)``)

    Examples
    --------
    >>> input_df = spark.createDataFrame([
    ...     Row(friends=[Row(first_name="Gianni", id=3993, last_name="Weber"),
    ...                  Row(first_name="Arielle", id=17484, last_name="Greaves")]),
    ... ])
    >>>
    >>> input_df.select(spq.as_is("friends.first_name")).show(truncate=False)
    +-----------------+
    |[Gianni, Arielle]|
    +-----------------+
    >>>
    >>> mapping = [("my_friends", "friends", spq.as_is)]
    >>> output_df = Mapper(mapping).transform(input_df)
    >>> output_df.show(truncate=False)
    +--------------------------------------------------+
    |my_friends                                        |
    +--------------------------------------------------+
    |[[Gianni, 3993, Weber], [Arielle, 17484, Greaves]]|
    +--------------------------------------------------+


    Returns
    -------
    partial or Column
        This method returns a suitable type depending on how it was called. This ensures compability
        with Spooq's mapper transformer - with or without explicit parameters - as well as direct calls via select,
        withColumn, where, ...
    """

    def _inner_func(source_column, name, alt_src_cols, cast):
        source_column = _coalesce_source_columns(source_column, alt_src_cols)
        if cast:
            source_column = source_column.try_cast(cast)
        return source_column.alias(name)

    args = dict(
        alt_src_cols=kwargs.get("alt_src_cols", False),
        cast=kwargs.get("cast", False),
    )

    return _get_executable_function(_inner_func, source_column, name, **args)


def to_num(source_column=None, name=None, **kwargs: Any) -> Union[partial, Column]:
    """
    More robust conversion to number data types (Default: LongType).
    This method is able to additionally handle (compared to implicit Spark conversion):

        * Preceding and/or trailing whitespace
        * underscores as 'thousand' separators

    Parameters
    ----------
    source_column : str or Column
        Input column. Can be a name, pyspark column or pyspark function
    name : str, default -> derived from input column
        Name of the output column. (``.alias(name)``)

    Keyword Arguments
    -----------------
    alt_src_cols : str, default -> no coalescing, only source_column
        Coalesce with source_column and columns from this parameter.
    cast : T.DataType(), default -> T.LongType()
        Applies provided datatype on output column (``.try_cast(cast)``)

    Examples
    --------
    >>> input_df = spark.createDataFrame(
    ...     [
    ...         Row(input_string="  123456 "),
    ...         Row(input_string="Hello"),
    ...         Row(input_string="123_456")
    ...     ], schema="input_key string"
    ... )
    >>>
    >>> input_df.select(spq.to_num("input_key")).show(truncate=False)
    +---------+
    |123456   |
    |null     |
    |123456   |
    +---------+
    >>>
    >>> mapping = [
    ...     ("original_value",    "input_key", spq.as_is),
    ...     ("transformed_value", "input_key", spq.to_num)
    ... ]
    >>> output_df = Mapper(mapping).transform(input_df)
    >>> output_df.show(truncate=False)
    +--------------+-----------------+
    |original_value|transformed_value|
    +--------------+-----------------+
    |  123456      |123456           |
    |Hello         |null             |
    |123_456       |123456           |
    +--------------+-----------------+

    Returns
    -------
    partial or Column
        This method returns a suitable type depending on how it was called. This ensures compability
        with Spooq's mapper transformer - with or without explicit parameters - as well as direct calls via select,
        withColumn, where, ...
    """

    def _inner_func(source_column, name, alt_src_cols, cast):
        source_column = _coalesce_source_columns(source_column, alt_src_cols)
        source_column = F.regexp_replace(F.trim(source_column), "_", "")

        if isinstance(cast, str):
            cast = T._parse_datatype_string(cast)

        if isinstance(cast, T.NumericType):
            if _datatype_is_of_integer_family(cast):
                source_column = F.round(source_column.try_cast(T.DoubleType()), 0)
            else:
                source_column = source_column.try_cast(T.DoubleType())

        return source_column.try_cast(cast).alias(name)

    args = dict(
        alt_src_cols=kwargs.get("alt_src_cols", False),
        cast=kwargs.get("cast", T.LongType()),
    )

    return _get_executable_function(_inner_func, source_column, name, **args)


def to_bool(source_column=None, name=None, **kwargs: Any) -> partial:
    """
    More robust conversion to BooleanType.
    This method is able to additionally handle (compared to implicit Spark conversion):

        * Preceding and/or trailing whitespace
        * Define additional strings for true/false values ("on"/"off", "enabled"/"disabled" are added by default)

    Parameters
    ----------
    source_column : str or Column
        Input column. Can be a name, pyspark column or pyspark function
    name : str, default -> derived from input column
        Name of the output column. (``.alias(name)``)

    Keyword Arguments
    -----------------
    case_sensitive : Bool, default -> False
        Defines whether the case for the additional true/false lookup values is considered
    true_values : list, default -> ["on", "enabled"]
        A list of values that should result in a ``True`` value if they are found in the source column
    false_values : list, default -> ["off", "disabled"]
        A list of values that should result in a ``False`` value if they are found in the source column
    replace_default_values : Bool, default -> False
        Defines whether additionally provided true/false values replace or extend the default list
    alt_src_cols : str, default -> no coalescing, only source_column
        Coalesce with source_column and columns from this parameter.
    cast : T.DataType(), default -> T.BooleanType()
        Applies provided datatype on output column (``.try_cast(cast)``)

    Warning
    ---------
    Spark (and Spooq) handles number to boolean conversions depending on the input datatype!
    Please see this table for clarification:

    +-------+----------+-----------------+-------------+
    | Input            | Result                        |
    +-------+----------+-----------------+-------------+
    | Value | Datatype | Cast to Boolean | spq.to_bool |
    +=======+==========+=================+=============+
    |  -1   | int      | True            | NULL        |
    +-------+----------+-----------------+-------------+
    |  -1   | str      | NULL            | NULL        |
    +-------+----------+-----------------+-------------+
    |   0   | int      | False           | False       |
    +-------+----------+-----------------+-------------+
    |   0   | str      | False           | False       |
    +-------+----------+-----------------+-------------+
    |   1   | int      | True            | True        |
    +-------+----------+-----------------+-------------+
    |   1   | str      | True            | True        |
    +-------+----------+-----------------+-------------+
    | 100   | int      | True            | NULL        |
    +-------+----------+-----------------+-------------+
    | 100   | str      | NULL            | NULL        |
    +-------+----------+-----------------+-------------+

    Examples
    --------
    >>> input_df = spark.createDataFrame(
    ...     [
    ...         Row(input_string="  false "),
    ...         Row(input_string="123"),
    ...         Row(input_string="1"),
    ...         Row(input_string="Enabled"),
    ...         Row(input_string="?"),
    ...         Row(input_string="n")
    ...     ], schema="input_key string"
    ... )
    >>>
    >>> input_df.select(spq.to_bool("input_key", false_values=["?"])).show(truncate=False)
    +---------+
    |false    |
    |null     |
    |true     |
    |false    |
    |false    |
    +---------+
    >>>
    >>> mapping = [
    ...     ("original_value",    "input_key", spq.as_is),
    ...     ("transformed_value", "input_key", spq.to_bool(false_values=["?"]))
    ... ]
    >>> output_df = Mapper(mapping).transform(input_df)
    >>> output_df.show(truncate=False)
    +--------------+-----------------+
    |original_value|transformed_value|
    +--------------+-----------------+
    |  false       |false            |
    |123           |null             |
    |1             |true             |
    |Enabled       |true             |
    |?             |false            |
    |n             |false            |
    +--------------+-----------------+

    Returns
    -------
    partial or Column
        This method returns a suitable type depending on how it was called. This ensures compability
        with Spooq's mapper transformer - with or without explicit parameters - as well as direct calls via select,
        withColumn, where, ...
    """

    def _inner_func(source_column, name, true_values, false_values, case_sensitive, alt_src_cols, cast):
        source_column = _coalesce_source_columns(source_column, alt_src_cols)
        if case_sensitive:
            true_condition = F.trim(source_column).isin(true_values)
            false_condition = F.trim(source_column).isin(false_values)
        else:
            true_values = [str(val).lower() for val in true_values]
            false_values = [str(val).lower() for val in false_values]
            true_condition = F.lower(F.trim(source_column)).isin(true_values)
            false_condition = F.lower(F.trim(source_column)).isin(false_values)

        return (
            (
                F.when(true_condition, F.lit(True))
                .when(false_condition, F.lit(False))
                .otherwise(F.trim(source_column).try_cast(T.BooleanType()))
            )
            .try_cast(cast)
            .alias(name)
        )

    args = dict(
        case_sensitive=kwargs.get("case_sensitive", False),
        alt_src_cols=kwargs.get("alt_src_cols", False),
        cast=kwargs.get("cast", T.BooleanType()),
    )

    if kwargs.get("replace_default_values"):
        args["true_values"] = kwargs.get("true_values", ["on", "enabled"])
        args["false_values"] = kwargs.get("false_values", ["off", "disabled"])
    else:
        args["true_values"] = kwargs.get("true_values", []) + ["on", "enabled"]
        args["false_values"] = kwargs.get("false_values", []) + ["off", "disabled"]

    return _get_executable_function(_inner_func, source_column, name, **args)


def to_timestamp(source_column=None, name=None, **kwargs: Any) -> partial:
    """
    More robust conversion to TimestampType (or as a formatted string).
    This method supports following input types:

        * Unix timestamps in seconds
        * Unix timestamps in milliseconds
        * Timestamps in any format supported by Spark
        * Timestamps in any custom format (via ``input_format``)
        * Preceding and/or trailing whitespace

    Parameters
    ----------
    source_column : str or Column
        Input column. Can be a name, pyspark column or pyspark function
    name : str, default -> derived from input column
        Name of the output column. (``.alias(name)``)

    Keyword Arguments
    -----------------
    max_timestamp_sec : int, default -> 4102358400 (=> 2099-12-31 01:00:00)
        Defines the range in which unix timestamps are still considered as seconds (compared to milliseconds)
    input_format : [str, Bool], default -> False
        Spooq tries to convert the input string with the provided pattern (via ``F.unix_timestamp()``)
    output_format : [str, Bool], default -> False
        The output can be formatted according to the provided pattern (via ``F.date_format()``)
    min_timestamp_ms : int, default -> -62135514321000 (=> Year 1)
        Defines the overall allowed range to keep the timestamps within Python's ``datetime`` library limits
    max_timestamp_ms : int, default -> 253402210800000 (=> Year 9999)
        Defines the overall allowed range to keep the timestamps within Python's ``datetime`` library limits
    alt_src_cols : str, default -> no coalescing, only source_column
        Coalesce with source_column and columns from this parameter.
    cast : T.DataType(), default -> T.TimestampType()
        Applies provided datatype on output column (``.try_cast(cast)``)

    Warning
    ---------
    * Timestamps in the range (-inf, -max_timestamp_sec) and (max_timestamp_sec, inf) are treated as milliseconds
    * There is a time interval (1970-01-01 +- ~2.5 months) where we can not distinguish correctly between s and ms
      (e.g. 3974400000 would be treated as seconds (2095-12-11T00:00:00) as the value is smaller than
      MAX_TIMESTAMP_S, but it could also be a valid date in Milliseconds (1970-02-16T00:00:00)

    Examples
    --------
    >>> input_df = spark.createDataFrame(
    ...     [
    ...         Row(input_key="2020-08-12T12:43:14+0000"),
    ...         Row(input_key="1597069446"),
    ...         Row(input_key="1597069446000"),
    ...         Row(input_key="2020-08-12"),
    ...     ], schema="input_key string"
    ... )
    >>>
    >>> input_df.select(spq.to_timestamp("input_key")).show(truncate=False)
    +-------------------+
    |2020-08-12 14:43:14|
    |2020-08-10 16:24:06|
    |2020-08-10 16:24:06|
    |2020-08-12 00:00:00|
    +-------------------+
    >>>
    >>> mapping = [
    ...     ("original_value",    "input_key", spq.as_is),
    ...     ("transformed_value", "input_key", spq.to_timestamp)
    ... ]
    >>> output_df = Mapper(mapping).transform(input_df)
    >>> output_df.show(truncate=False)
    +------------------------+-------------------+
    |original_value          |transformed_value  |
    +------------------------+-------------------+
    |2020-08-12T12:43:14+0000|2020-08-12 14:43:14|
    |1597069446              |2020-08-10 16:24:06|
    |1597069446000           |2020-08-10 16:24:06|
    |2020-08-12              |2020-08-12 00:00:00|
    +------------------------+-------------------+

    Returns
    -------
    partial or Column
        This method returns a suitable type depending on how it was called. This ensures compability
        with Spooq's mapper transformer - with or without explicit parameters - as well as direct calls via select,
        withColumn, where, ...
    """

    def _inner_func(
        source_column,
        name,
        max_timestamp_sec,
        input_format,
        output_format,
        min_timestamp_ms,
        max_timestamp_ms,
        alt_src_cols,
        cast,
    ):
        source_column = _coalesce_source_columns(source_column, alt_src_cols)

        if input_format:
            output_col = (
                F.try_to_timestamp(source_column.try_cast(T.StringType()), F.lit(input_format)).alias(name)
            )
        else:
            output_col = (
                F.when(
                    ~F.trim(source_column).try_cast(T.LongType()).between(min_timestamp_ms, max_timestamp_ms),
                    F.lit(None),
                )
                .when(
                    F.abs(F.trim(source_column).try_cast(T.LongType())).between(0, max_timestamp_sec),
                    F.try_to_timestamp(F.trim(source_column).try_cast(T.LongType())),
                )
                .when(
                    F.abs(F.trim(source_column).try_cast(T.LongType())) > max_timestamp_sec,
                    F.try_to_timestamp(F.trim(source_column) / 1000),
                )
                .otherwise(F.try_to_timestamp(F.trim(source_column)))
            )
        if output_format:
            output_col = F.date_format(output_col, output_format)
            cast = T.StringType()
        return output_col.try_cast(cast).alias(name)

    args = dict(
        max_timestamp_sec=kwargs.get("max_timestamp_sec", 4102358400),
        input_format=kwargs.get("input_format", False),
        output_format=kwargs.get("output_format", False),
        min_timestamp_ms=kwargs.get("min_timestamp_ms", -62135514321000),
        max_timestamp_ms=kwargs.get("max_timestamp_ms", 253402210800000),
        alt_src_cols=kwargs.get("alt_src_cols", False),
        cast=kwargs.get("cast", T.TimestampType()),
    )

    return _get_executable_function(_inner_func, source_column, name, **args)


def str_to_array(source_column=None, name=None, **kwargs: Any) -> partial:
    """
    Splits a string into a list (ArrayType). Supports leading/trailing whitespace, square brackets
        and custom separators.

    Parameters
    ----------
    source_column : str or Column
        Input column. Can be a name, pyspark column or pyspark function
    name : str, default -> derived from input column
        Name of the output column. (``.alias(name)``)

    Keyword Arguments
    -----------------
    sep : str, default -> ","
        Separator that is used to split the input string.
    alt_src_cols : str, default -> no coalescing, only source_column
        Coalesce with source_column and columns from this parameter.
    cast : T.DataType(), default -> T.StringType()
        Applies provided datatype on the elements of the output array (``.try_cast(T.ArrayType(cast))``)

    Examples
    --------
    >>> input_df = spark.createDataFrame(
    ...     [
    ...         Row(input_key="[item1,item2,3]"),
    ...         Row(input_key="item1,it[e]m2,it em3"),
    ...         Row(input_key="    item1,   item2    ,   item3")
    ...     ], schema="input_key string"
    ... )
    >>>
    >>> input_df.select(spq.str_to_array("input_key")).show(truncate=False)
    +------------------------+
    |[item1, item2, 3]       |
    |[item1, it[e]m2, it em3]|
    |[item1, item2, item3]   |
    +------------------------+
    >>>
    >>> mapping = [
    ...     ("original_value",    "input_key", spq.as_is),
    ...     ("transformed_value", "input_key", spq.str_to_array)
    ... ]
    >>> output_df = Mapper(mapping).transform(input_df)
    >>> output_df.printSchema()
    root
     |-- original_value: string (nullable = true)
     |-- transformed_value: array (nullable = true)
     |    |-- element: string (containsNull = true)
    >>> output_df.show(truncate=False)
    +-------------------------------+------------------------+
    |original_value                 |transformed_value       |
    +-------------------------------+------------------------+
    |[item1,item2,3]                |[item1, item2, 3]       |
    |item1,it[e]m2,it em3           |[item1, it[e]m2, it em3]|
    |    item1,   item2    ,   item3|[item1, item2, item3]   |
    +-------------------------------+------------------------+

    Returns
    -------
    partial or Column
        This method returns a suitable type depending on how it was called. This ensures compability
        with Spooq's mapper transformer - with or without explicit parameters - as well as direct calls via select,
        withColumn, where, ...
    """

    def _inner_func(source_column, name, sep, alt_src_cols, cast):
        source_column = _coalesce_source_columns(source_column, alt_src_cols)

        if isinstance(cast, str):
            cast = T._parse_datatype_string(cast)

        if isinstance(cast, T.NumericType):
            if _datatype_is_of_integer_family(cast):
                transform_func = lambda x: F.round(x.try_cast(T.DoubleType()), 0)
            else:
                transform_func = lambda x: x.try_cast(T.DoubleType())
        else:
            transform_func = lambda x: x

        output_type = T.ArrayType(cast)

        return (
            F.transform(
                F.split(F.regexp_replace(source_column, r"^\s*\[*\s*|\s*\]*\s*$", ""), rf"\s*{sep}\s*"),
                transform_func,
            )
            .try_cast(output_type)
            .alias(name)
        )

    args = dict(
        sep=kwargs.get("sep", ","),
        alt_src_cols=kwargs.get("alt_src_cols", False),
        cast=kwargs.get("cast", T.StringType()),
    )

    return _get_executable_function(_inner_func, source_column, name, **args)


def map_values(source_column=None, name=None, **kwargs: Any) -> partial:
    """
    Maps input values to specified output values.

    Parameters
    ----------
    source_column : str or Column
        Input column. Can be a name, pyspark column or pyspark function
    name : str, default -> derived from input column
        Name of the output column. (``.alias(name)``)

    Keyword Arguments
    -----------------
    mapping : dict
        Dictionary containing lookup / substitute value pairs.
    default : [str, Column, Any], default -> "source_column"
        Defines what will be returned if no matching lookup value was found.
    ignore_case : bool, default -> True
        Only relevant for "equals" and "sql_like" comparison operators.
    pattern_type : str, default -> "equals"
        Please choose among ['equals', 'regex' and 'sql_like'] for the comparison of input value and mapping key.
    alt_src_cols : str, default -> no coalescing, only source_column
        Coalesce with source_column and columns from this parameter.
    cast : T.DataType(), default -> T.StringType()
        Applies provided datatype on output column (``.try_cast(cast)``)

    Hint
    ----
    Maybe this table helps you to better understand what happens behind the curtains:

    +---------------+------------+---------------------------------------------------------------------------+
    | lookup        | substitute | mode         | Internal Spark Logic                                       |
    +===============+============+=====================================+==================+==================+
    | whitelist     | allowlist  | "equals"     | F.when(                                                    |
    |               |            |              |     F.col("input_column") == "whitelist",                  |
    |               |            |              |     F.lit("allowlist")                                     |
    |               |            |              | ).otherwise(                                               |
    |               |            |              |     F.col("input_column")                                  |
    |               |            |              | )                                                          |
    +---------------+------------+---------------------------------------------------------------------------+
    | %whitelist%   | allowlist  | "sql_like"   | F.when(                                                    |
    |               |            |              |     F.col("input_column").like("%whitelist%",              |
    |               |            |              |     F.lit("allowlist")                                     |
    |               |            |              | ).otherwise(                                               |
    |               |            |              |     F.col("input_column")                                  |
    |               |            |              | )                                                          |
    +---------------+------------+---------------------------------------------------------------------------+
    | .*whitelist.* | allowlist  | "regex   "   | F.when(                                                    |
    |               |            |              |     F.col("input_column").rlike(".*whitelist.*",           |
    |               |            |              |     F.lit("allowlist")                                     |
    |               |            |              | ).otherwise(                                               |
    |               |            |              |     F.col("input_column")                                  |
    |               |            |              | )                                                          |
    +---------------+------------+---------------------------------------------------------------------------+

    Examples
    --------
    >>> input_df = spark.createDataFrame(
    ...     [
    ...         ("allowlist", ),
    ...         ("WhiteList", ),
    ...         ("blocklist", ),
    ...         ("blacklist", ),
    ...         ("Blacklist", ),
    ...         ("Shoppinglist", ),
    ...     ], schema="input_key string"
    ... )
    >>> substitute_mapping = {"whitelist": "allowlist", "blacklist": "blocklist"}
    >>>
    >>> input_df.select(spq.map_values("input_key", mapping=substitute_mapping)).show(truncate=False)
    +------------+
    |allowlist   |
    |allowlist   |
    |blocklist   |
    |blocklist   |
    |blocklist   |
    |Shoppinglist|
    +------------+
    >>>
    >>> mapping = [
    ...     ("original_value",    "input_key", spq.as_is),
    ...     ("transformed_value", "input_key", spq.map_values(mapping=substitute_mapping))
    ... ]
    >>> output_df = Mapper(mapping).transform(input_df)
    >>> output_df.show(truncate=False)
    +--------------+-----------------+
    |original_value|transformed_value|
    +--------------+-----------------+
    |allowlist     |allowlist        |
    |WhiteList     |allowlist        |
    |blocklist     |blocklist        |
    |blacklist     |blocklist        |
    |Blacklist     |blocklist        |
    |Shoppinglist  |Shoppinglist     |
    +--------------+-----------------+

    Returns
    -------
    partial or Column
        This method returns a suitable type depending on how it was called. This ensures compability
        with Spooq's mapper transformer - with or without explicit parameters - as well as direct calls via select,
        withColumn, where, ...
    """

    def _inner_func(source_column, name, mapping, default, ignore_case, pattern_type, alt_src_cols, cast):
        source_column = _coalesce_source_columns(source_column, alt_src_cols)

        if isinstance(default, str) and default == "source_column":
            default = source_column
        elif not isinstance(default, Column):
            default = F.lit(default)

        if ignore_case and pattern_type != "regex":
            mapping = {str(key).lower(): F.lit(value).try_cast(cast) for key, value in mapping.items()}
            source_column = F.lower(source_column)
        else:
            mapping = {key: F.lit(value).try_cast(cast) for key, value in mapping.items()}

        keys = list(mapping.keys())
        if pattern_type == "equals":
            when_clause = F.when(source_column.cast(T.StringType()) == str(keys[0]), mapping[keys[0]])
            for key in keys[1:]:
                when_clause = when_clause.when(source_column.try_cast(T.StringType()) == str(key), mapping[key])

        elif pattern_type == "sql_like":
            when_clause = F.when(source_column.cast(T.StringType()).like(keys[0]), mapping[keys[0]])
            for key in keys[1:]:
                when_clause = when_clause.when(source_column.cast(T.StringType()).like(key), mapping[key])

        elif pattern_type == "regex":
            when_clause = F.when(source_column.cast(T.StringType()).rlike(keys[0]), mapping[keys[0]])
            for key in keys[1:]:
                when_clause = when_clause.when(source_column.cast(T.StringType()).rlike(key), mapping[key])

        else:
            raise ValueError(
                f"pattern_type <{pattern_type}> not recognized. "
                "Please choose among ['equals' (default), 'regex' and 'sql_like']"
            )

        when_clause = when_clause.otherwise(default.try_cast(cast))

        return when_clause.alias(name)

    try:
        mapping = kwargs["mapping"]
    except TypeError:
        raise TypeError("'map_values' is missing the mapping dict (f.e. mapping=dict(results='training'))")
    if len(mapping.keys()) < 1:
        raise ValueError("'map_values' received an empty map (f.e. mapping=dict(results='training'))")

    args = dict(
        mapping=mapping,
        default=kwargs.get("default", "source_column"),
        ignore_case=kwargs.get("ignore_case", True),
        pattern_type=kwargs.get("pattern_type", "equals"),
        alt_src_cols=kwargs.get("alt_src_cols", False),
        cast=kwargs.get("cast", T.StringType()),
    )

    return _get_executable_function(_inner_func, source_column, name, **args)


def meters_to_cm(source_column=None, name=None, **kwargs: Any) -> partial:
    """
    Converts meters to cm and casts the result to an IntegerType.

    Parameters
    ----------
    source_column : str or Column
        Input column. Can be a name, pyspark column or pyspark function
    name : str, default -> derived from input column
        Name of the output column. (``.alias(name)``)

    Keyword Arguments
    -----------------
    alt_src_cols : str, default -> no coalescing, only source_column
        Coalesce with source_column and columns from this parameter.
    cast : T.DataType(), default -> T.IntegerType()
        Applies provided datatype on output column (``.try_cast(cast)``)

    Examples
    --------
    >>> input_df = spark.createDataFrame([
    ...     Row(size_in_m=1.80),
    ...     Row(size_in_m=1.65),
    ...     Row(size_in_m=2.05)
    ... ])
    >>>
    >>> input_df.select(spq.meters_to_cm("size_in_m")).show(truncate=False)
    +---------+
    |180      |
    |165      |
    |204      |
    +---------+
    >>>
    >>> mapping = [
    ...     ("original_value",  "size_in_m",  spq.as_is),
    ...     ("size_in_cm",      "size_in_m",  spq.meters_to_cm),
    ... ]
    >>> output_df = Mapper(mapping).transform(input_df)
    >>> output_df.show(truncate=False)
    +--------------+----------+
    |original_value|size_in_cm|
    +--------------+----------+
    |1.8           |180       |
    |1.65          |165       |
    |2.05          |204       |
    +--------------+----------+

    Returns
    -------
    partial or Column
        This method returns a suitable type depending on how it was called. This ensures compability
        with Spooq's mapper transformer - with or without explicit parameters - as well as direct calls via select,
        withColumn, where, ...
    """

    def _inner_func(source_column, name, alt_src_cols, cast):
        source_column = _coalesce_source_columns(source_column, alt_src_cols)
        if not isinstance(cast, str):
            cast = cast.simpleString()
        if _datatype_is_of_integer_family(cast):
            source_column_calculated = F.round(source_column.try_cast(T.DoubleType()) * 100, F.lit(0))
        else:
            source_column_calculated = source_column.try_cast(T.DoubleType()) * 100
        return source_column_calculated.try_cast(cast).alias(name)

    args = dict(
        alt_src_cols=kwargs.get("alt_src_cols", False),
        cast=kwargs.get("cast", T.IntegerType()),
    )

    return _get_executable_function(_inner_func, source_column, name, **args)


def has_value(source_column=None, name=None, **kwargs: Any) -> partial:
    """
    Returns True if the source_column is
        - not NULL and
        - not "" (empty string)
        - otherwise it returns False

    Warning
    -------
    This means that it will return True for values which would indicate a False value. Like "false" or 0!!!

    Parameters
    ----------
    source_column : str or Column
        Input column. Can be a name, pyspark column or pyspark function
    name : str, default -> derived from input column
        Name of the output column. (``.alias(name)``)

    Keyword Arguments
    -----------------
    alt_src_cols : str, default -> no coalescing, only source_column
        Coalesce with source_column and columns from this parameter.
    cast : T.DataType(), default -> T.BooleanType()
        Applies provided datatype on output column (``.try_cast(cast)``)

    Examples
    --------
    >>> input_df = spark.createDataFrame(
    ...     [
    ...         Row(input_key=False),
    ...         Row(input_key=None),
    ...         Row(input_key="some text"),
    ...         Row(input_key="")
    ...     ], schema="input_key string"
    ... )
    >>>
    >>> input_df.select(spq.has_value("input_key")).show(truncate=False)
    +---------+
    |true     |
    |false    |
    |true     |
    |false    |
    +---------+
    >>>
    >>> mapping = [
    ...     ("original_value",      "input_key",  spq.as_is),
    ...     ("has_value",  "input_key",  spq.has_value)
    ... ]
    >>> output_df = Mapper(mapping).transform(input_df)
    >>> output_df.show(truncate=False)
    +--------------+---------+
    |original_value|has_value|
    +--------------+---------+
    |false         |true     |
    |null          |false    |
    |some text     |true     |
    |              |false    |
    +--------------+---------+

    Returns
    -------
    partial or Column
        This method returns a suitable type depending on how it was called. This ensures compability
        with Spooq's mapper transformer - with or without explicit parameters - as well as direct calls via select,
        withColumn, where, ...
    """

    def _inner_func(source_column, name, alt_src_cols, cast):
        source_column = _coalesce_source_columns(source_column, alt_src_cols)
        return (
            F.when((source_column.isNotNull()) & (source_column.try_cast(T.StringType()) != ""), F.lit(True))
            .otherwise(F.lit(False))
            .try_cast(cast)
            .alias(name)
        )

    args = dict(
        alt_src_cols=kwargs.get("alt_src_cols", False),
        cast=kwargs.get("cast", T.BooleanType()),
    )

    return _get_executable_function(_inner_func, source_column, name, **args)


def apply(source_column=None, name=None, **kwargs: Any) -> partial:
    """
    Applies a function / partial

    Parameters
    ----------
    source_column : str or Column
        Input column. Can be a name, pyspark column or pyspark function
    name : str, default -> derived from input column
        Name of the output column. (``.alias(name)``)

    Keyword Arguments
    -----------------
    func : Callable
        Function that takes the source column as single argument
    alt_src_cols : str, default -> no coalescing, only source_column
        Coalesce with source_column and columns from this parameter.
    cast : T.DataType(), default -> no casting
        Applies provided datatype on output column (``.try_cast(cast)``)

    Examples
    --------
    >>> input_df = spark.createDataFrame(
    ...     [
    ...         ("F", ),
    ...         ("f", ),
    ...         ("x", ),
    ...         ("X", ),
    ...         ("m", ),
    ...         ("M", ),
    ...     ], schema="input_key string"
    ... )
    >>>
    >>> input_df.select(spq.apply("input_key", func=F.lower)).show(truncate=False)
    +---------+
    |f        |
    |f        |
    |x        |
    |x        |
    |m        |
    |m        |
    +---------+
    >>>
    >>> mapping = [
    ...     ("original_value",    "input_key", spq.as_is),
    ...     ("transformed_value", "input_key", spq.apply(func=F.lower))
    ... ]
    >>> output_df = Mapper(mapping).transform(input_df)
    >>> output_df.show(truncate=False)
    +--------------+-----------------+
    |original_value|transformed_value|
    +--------------+-----------------+
    |F             |f                |
    |f             |f                |
    |x             |x                |
    |X             |x                |
    |m             |m                |
    |M             |m                |
    +--------------+-----------------+

    >>> input_df = spark.createDataFrame(
    ...     [
    ...         ("sarajishvilileqso@gmx.at", ),
    ...         ("jnnqn@astrinurdin.art", ),
    ...         ("321aw@hotmail.com", ),
    ...         ("techbrenda@hotmail.com", ),
    ...         ("sdsxcx@gmail.com", ),
    ...     ], schema="input_key string"
    ... )
    ...
    >>> def _has_hotmail(source_column):
    ...     return F.when(
    ...         source_column.try_cast(T.StringType()).endswith("@hotmail.com"),
    ...         F.lit(True)
    ...     ).otherwise(F.lit(False))
    ...
    >>> mapping = [
    ...     ("original_value", "input_key", spq.as_is),
    ...     ("over_sixty",     "input_key", spq.apply(func=_has_hotmail))
    ... ]
    >>> output_df = Mapper(mapping).transform(input_df)
    >>> output_df.show(truncate=False)
    +------------------------+----------+
    |original_value          |over_sixty|
    +------------------------+----------+
    |sarajishvilileqso@gmx.at|false     |
    |jnnqn@astrinurdin.art   |false     |
    |321aw@hotmail.com       |true      |
    |techbrenda@hotmail.com  |true      |
    |sdsxcx@gmail.com        |false     |
    +------------------------+----------+

    Returns
    -------
    partial or Column
        This method returns a suitable type depending on how it was called. This ensures compability
        with Spooq's mapper transformer - with or without explicit parameters - as well as direct calls via select,
        withColumn, where, ...
    """

    def _inner_func(source_column, name, func, alt_src_cols, cast):
        source_column = _coalesce_source_columns(source_column, alt_src_cols)
        source_column = func(source_column)
        if cast:
            source_column = source_column.try_cast(cast)
        return source_column.alias(name)

    try:
        func = kwargs["func"]
    except TypeError:
        raise TypeError("'apply' transformation is missing the custom function (f.e. func=F.lower)")

    args = dict(
        func=func,
        alt_src_cols=kwargs.get("alt_src_cols", False),
        cast=kwargs.get("cast", False),
    )

    return _get_executable_function(_inner_func, source_column, name, **args)


def to_json_string(source_column=None, name=None, **kwargs: Any) -> partial:
    """
    Returns a column as json compatible string. Nested hierarchies are supported.
    This function also supports NULL and strings as input in comparison to Spark's built-in ``to_json``.
    The unicode representation of a column will be returned if an error occurs.

    Parameters
    ----------
    source_column : str or Column
        Input column. Can be a name, pyspark column or pyspark function
    name : str, default -> derived from input column
        Name of the output column. (``.alias(name)``)

    Keyword Arguments
    -----------------
    alt_src_cols : str, default -> no coalescing, only source_column
        Coalesce with source_column and columns from this parameter.
    cast : T.DataType(), default -> no casting, same return data type as input data type
        Applies provided datatype on output column (``.try_cast(cast)``)

    Examples
    --------
    >>> input_df = spark.createDataFrame([
    ...     Row(friends=[Row(first_name="Gianni", id=3993, last_name="Weber"),
    ...                  Row(first_name="Arielle", id=17484, last_name="Greaves")]),
    ... ])
    >>>
    >>> input_df.select(spq.to_json_string("friends")).show(truncate=False)
    +----------------------------------------------------------------------------------------------------------------------------+
    |[{"first_name": "Gianni", "id": 3993, "last_name": "Weber"}, {"first_name": "Arielle", "id": 17484, "last_name": "Greaves"}]|
    +----------------------------------------------------------------------------------------------------------------------------+
    >>>
    >>> mapping = [("friends_json", "friends", spq.to_json_string)]
    >>> output_df = Mapper(mapping).transform(input_df)
    >>> output_df.show(truncate=False)
    +----------------------------------------------------------------------------------------------------------------------------+
    |friends_json                                                                                                                |
    +----------------------------------------------------------------------------------------------------------------------------+
    |[{"first_name": "Gianni", "id": 3993, "last_name": "Weber"}, {"first_name": "Arielle", "id": 17484, "last_name": "Greaves"}]|
    +----------------------------------------------------------------------------------------------------------------------------+
    >>> input_df.select(spq.to_json_string("friends.first_name")).show(truncate=False)
    +---------------------+
    |first_name           |
    +---------------------+
    |['Gianni', 'Arielle']|
    +---------------------+

    Returns
    -------
    partial or Column
        This method returns a suitable type depending on how you called it. This ensures compability
        with Spooq's mapper transformer - with or without explicit parameters as well as direct calls via select,
        withColumn, where, ...
    """

    def _inner_func(source_column, name, alt_src_cols, cast):
        def _to_json(col):
            if not col:
                return None
            try:
                if isinstance(col, list):
                    return json.dumps([x.asDict(recursive=True) for x in col])
                else:
                    return json.dumps(col.asDict(recursive=True))
            except (AttributeError, TypeError):
                return str(col)

        source_column = _coalesce_source_columns(source_column, alt_src_cols)

        udf_to_json = F.udf(_to_json, cast)
        return udf_to_json(source_column).alias(name)

    args = dict(
        alt_src_cols=kwargs.get("alt_src_cols", False),
        cast=kwargs.get("cast", T.StringType()),
    )

    return _get_executable_function(_inner_func, source_column, name, **args)


# Convenience Methods


def to_str(source_column=None, name=None, **kwargs: Any) -> Union[partial, Column]:
    """
    Convenience transformation that only casts to string.

    Parameters
    ----------
    source_column : str or Column
        Input column. Can be a name, pyspark column or pyspark function
    name : str, default -> derived from input column
        Name of the output column. (``.alias(name)``)

    Keyword Arguments
    -----------------
    alt_src_cols : str, default -> no coalescing, only source_column
        Coalesce with source_column and columns from this parameter.

    Examples
    --------
    >>> input_df = spark.createDataFrame(
    ...     [
    ...         Row(input_key=123456),
    ...         Row(input_key=-123456),
    ...     ], schema="input_key int"
    ... )
    >>>
    >>> input_df.select(spq.to_str("input_key")).show(truncate=False)
    +---------+
    |123456   |
    |-123456  |
    +---------+
    >>>
    >>> mapping = [
    ...     ("original_value",    "input_key", spq.as_is),
    ...     ("transformed_value", "input_key", spq.to_str)
    ... ]
    >>> output_df = Mapper(mapping).transform(input_df)
    >>> output_df.show(truncate=False)
    +--------------+-----------------+
    |original_value|transformed_value|
    +--------------+-----------------+
    |123456        |123456           |
    |-123456       |-123456          |
    +--------------+-----------------+

    Returns
    -------
    partial or Column
        This method returns a suitable type depending on how it was called. This ensures compability
        with Spooq's mapper transformer - with or without explicit parameters - as well as direct calls via select,
        withColumn, where, ...
    """

    def _inner_func(source_column, name, alt_src_cols):
        source_column = _coalesce_source_columns(source_column, alt_src_cols)
        return source_column.try_cast(T.StringType()).alias(name)

    args = dict(
        alt_src_cols=kwargs.get("alt_src_cols", False),
    )

    return _get_executable_function(_inner_func, source_column, name, **args)


def to_int(source_column=None, name=None, **kwargs: Any) -> Union[partial, Column]:
    """
    Syntactic sugar for calling ``to_num(cast="int")``

    Parameters
    ----------
    source_column : str or Column
        Input column. Can be a name, pyspark column or pyspark function
    name : str, default -> derived from input column
        Name of the output column. (``.alias(name)``)

    Keyword Arguments
    -----------------
    alt_src_cols : str, default -> no coalescing, only source_column
        Coalesce with source_column and columns from this parameter.

    Returns
    -------
    partial or Column
        This method returns a suitable type depending on how it was called. This ensures compability
        with Spooq's mapper transformer - with or without explicit parameters - as well as direct calls via select,
        withColumn, where, ...
    """

    args = dict(
        alt_src_cols=kwargs.get("alt_src_cols", False),
        cast="int",
    )

    return to_num(source_column, name, **args)


def to_long(source_column=None, name=None, **kwargs: Any) -> Union[partial, Column]:
    """
    Syntactic sugar for calling ``to_num(cast="long")``

    Parameters
    ----------
    source_column : str or Column
        Input column. Can be a name, pyspark column or pyspark function
    name : str, default -> derived from input column
        Name of the output column. (``.alias(name)``)

    Keyword Arguments
    -----------------
    alt_src_cols : str, default -> no coalescing, only source_column
        Coalesce with source_column and columns from this parameter.

    Returns
    -------
    partial or Column
        This method returns a suitable type depending on how it was called. This ensures compability
        with Spooq's mapper transformer - with or without explicit parameters - as well as direct calls via select,
        withColumn, where, ...
    """

    args = dict(
        alt_src_cols=kwargs.get("alt_src_cols", False),
        cast="long",
    )

    return to_num(source_column, name, **args)


def to_float(source_column=None, name=None, **kwargs: Any) -> Union[partial, Column]:
    """
    Syntactic sugar for calling ``to_num(cast="float")``

    Parameters
    ----------
    source_column : str or Column
        Input column. Can be a name, pyspark column or pyspark function
    name : str, default -> derived from input column
        Name of the output column. (``.alias(name)``)

    Keyword Arguments
    -----------------
    alt_src_cols : str, default -> no coalescing, only source_column
        Coalesce with source_column and columns from this parameter.

    Returns
    -------
    partial or Column
        This method returns a suitable type depending on how it was called. This ensures compability
        with Spooq's mapper transformer - with or without explicit parameters - as well as direct calls via select,
        withColumn, where, ...
    """

    args = dict(
        alt_src_cols=kwargs.get("alt_src_cols", False),
        cast="float",
    )

    return to_num(source_column, name, **args)


def to_double(source_column=None, name=None, **kwargs: Any) -> Union[partial, Column]:
    """
    Syntactic sugar for calling ``to_num(cast="double")``

    Parameters
    ----------
    source_column : str or Column
        Input column. Can be a name, pyspark column or pyspark function
    name : str, default -> derived from input column
        Name of the output column. (``.alias(name)``)

    Keyword Arguments
    -----------------
    alt_src_cols : str, default -> no coalescing, only source_column
        Coalesce with source_column and columns from this parameter.

    Returns
    -------
    partial or Column
        This method returns a suitable type depending on how it was called. This ensures compability
        with Spooq's mapper transformer - with or without explicit parameters - as well as direct calls via select,
        withColumn, where, ...
    """

    args = dict(
        alt_src_cols=kwargs.get("alt_src_cols", False),
        cast="double",
    )

    return to_num(source_column, name, **args)



from __future__ import absolute_import
from pyspark.sql import functions as f

from .transformer import Transformer


class Exploder(Transformer):
    """
    Explodes an array within a DataFrame and
    drops the column containing the source array.

    Examples
    --------
    >>> transformer = Exploder(
    >>>     path_to_array="attributes.friends",
    >>>     exploded_elem_name="friend",
    >>> )

    Parameters
    ----------
    path_to_array : :any:`str`, (Defaults to 'included')
        Defines the Column Name / Path to the Array.
        Dropping nested columns is not supported.
        Although, you can still explode them.

    exploded_elem_name : :any:`str`, (Defaults to 'elem')
        Defines the column name the exploded column will get.
        This is important to know how to access the Field afterwards.
        Writing nested columns is not supported.
        The output column has to be first level.

    drop_rows_with_empty_array : :any:`bool`, (Defaults to True)
        By default Spark (and Spooq) drops rows which don't have any elements in the array
        which is being exploded. To work-around this, set `drop_rows_with_empty_array` to False.

    Warning
    -------
    **Support for nested column:**

    path_to_array:
        PySpark cannot drop a field within a struct. This means the specific field
        can be referenced and therefore exploded, but not dropped.
    exploded_elem_name:
        If you (re)name a column in the dot notation, is creates a first level column,
        just with a dot its name. To create a struct with the column as a field
        you have to redefine the structure or use a UDF.


    Note
    ----
    The :meth:`~spark.sql.functions.explode` or :meth:`~spark.sql.functions.explode_outer` methods
    of Spark are used internally, depending on the `drop_rows_with_empty_array` parameter.

    Note
    ----
    The size of the resulting DataFrame is not guaranteed to be
    equal to the Input DataFrame!
    """

    def __init__(self, path_to_array="included", exploded_elem_name="elem", drop_rows_with_empty_array=True):
        super(Exploder, self).__init__()
        self.path_to_array = path_to_array
        self.exploded_elem_name = exploded_elem_name
        self.drop_rows_with_empty_array = drop_rows_with_empty_array

    def transform(self, input_df):
        if self.drop_rows_with_empty_array:
            explode_function = f.explode
        else:
            explode_function = f.explode_outer

        return input_df.withColumn(self.exploded_elem_name, explode_function(self.path_to_array)).drop(
            self.path_to_array
        )



import sys
import pyspark.sql.functions as F, types as T
from pyspark.sql.column import Column

from .base_cleaner import BaseCleaner


class EnumCleaner(BaseCleaner):
    """
    Cleanses a dataframe based on lists of allowed|disallowed values.

    Examples
    --------
    >>> from spooq.transformer import EnumCleaner
    >>>
    >>> transformer = EnumCleaner(
    >>>     cleaning_definitions={
    >>>         "status": {
    >>>             "elements": ["active", "inactive"],
    >>>         },
    >>>         "version": {
    >>>             "elements": ["", "None", "none", "null", "NULL"],
    >>>             "mode": "disallow",
    >>>             "default": None,
    >>>         },
    >>>     }
    >>> )

    >>> from spooq.transformer import EnumCleaner
    >>> from pyspark.sql import Row
    >>>
    >>> input_df = spark.createDataFrame([
    >>>     Row(a="stay", b="positive"),
    >>>     Row(a="stay", b="negative"),
    >>>     Row(a="stay", b="positive"),
    >>> ])
    >>> transformer = EnumCleaner(
    >>>     cleaning_definitions={
    >>>         "b": {
    >>>             "elements": ["positive"],
    >>>             "mode": "allow",
    >>>         }
    >>>     },
    >>>     column_to_log_cleansed_values="cleansed_values_enum",
    >>>     store_as_map=True,
    >>> )
    >>> output_df = transformer.transform(input_df)
    >>> output_df.show()
    +----+--------+--------------------+
    |   a|       b|cleansed_values_enum|
    +----+--------+--------------------+
    |stay|positive|                  []|
    |stay|    null|     [b -> negative]|
    |stay|positive|                  []|
    +----+--------+--------------------+
    >>> output_df.printSchema()
    root
     |-- a: string (nullable = true)
     |-- b: string (nullable = true)
     |-- cleansed_values_enum: map (nullable = false)
     |    |-- key: string
     |    |-- value: string (valueContainsNull = true)

    Parameters
    ----------
    cleaning_definitions : :py:class:`dict`
        Dictionary containing column names and respective cleansing rules

    column_to_log_cleansed_values : :any:`str`, Defaults to None
        Defines a column in which the original (uncleansed) value will be stored in case of cleansing. If no column
        name is given, nothing will be logged.

    store_as_map : :any:`bool`, Defaults to False
        Specifies if the logged cleansed values should be stored in a column as :any:`pyspark.sql.types.MapType` with
        stringified values or as :any:`pyspark.sql.types.StructType` with the original respective data types.

    Note
    ----
    Following cleansing rule attributes per column are supported:

        * elements, mandatory - :class:`list`
            A list of elements which will be used to allow or reject (based on mode) values from the input DataFrame.
        * mode, allow|disallow, defaults to 'allow' - :any:`str`
            "allow" will set all values which are NOT in the list (ignoring NULL) to the default value.
            "disallow" will set all values which ARE in the list (ignoring NULL) to the default value.
        * default, defaults to None - :class:`~pyspark.sql.column.Column` or any primitive Python value
            If a value gets cleansed it gets replaced with the provided default value.

    Returns
    -------
    |SPARK_DATAFRAME|
        The transformed DataFrame

    Raises
    ------
    :any:`exceptions.ValueError`
        Enumeration-based cleaning requires a non-empty list of elements per cleaning rule!
        Spooq did not find such a list for column: {column_name}

    :any:`exceptions.ValueError`
        Only the following modes are supported by EnumCleaner: 'allow' and 'disallow'.

    Warning
    -------
    None values are explicitly ignored as input values because `F.lit(None).isin(["elem1", "elem2"])` will neither
    return True nor False but None.
    If you want to replace Null values you should use the transformer :py:class:`spooq.transformer.NullCleaner`
    """

    def __init__(self, cleaning_definitions={}, column_to_log_cleansed_values=None, store_as_map=False):
        super().__init__(
            cleaning_definitions=cleaning_definitions,
            column_to_log_cleansed_values=column_to_log_cleansed_values,
            store_as_map=store_as_map,
            temporary_columns_prefix="9b7798529fef529c8f2586be7ca43a66",
        )
        self.logger.debug("Enumeration List: " + str(self.cleaning_definitions))

    def transform(self, input_df):
        self.logger.debug("input_df Schema: " + input_df._jdf.schema().treeString())
        if self.column_to_log_cleansed_values:
            column_names_to_clean = self.cleaning_definitions.keys()
            temporary_column_names = self._get_temporary_column_names(column_names_to_clean)
            input_df = self._add_temporary_columns(input_df, column_names_to_clean, temporary_column_names)

        for column_name, cleaning_definition in list(self.cleaning_definitions.items()):
            self.logger.debug(f"Cleaning Definition for Column {column_name}: {str(cleaning_definition)}")
            elements = cleaning_definition.get("elements", None)
            if not elements:
                raise ValueError(
                    f"Enumeration-based cleaning requires a non-empty list of elements per cleaning rule!",
                    f"\nSpooq did not find such a list for column: {column_name}",
                )
            mode = cleaning_definition.get("mode", "NOT_DEFINED")
            substitute = cleaning_definition.get("default", None)
            data_type = input_df.schema[column_name].dataType
            if not isinstance(substitute, Column):
                substitute = F.lit(substitute)

            if mode == "allow":
                input_df = input_df.withColumn(
                    column_name,
                    F.when(
                        F.col(column_name).isNull(),
                        F.lit(None)
                    ).otherwise(
                        F.when(
                            F.col(column_name).isin(elements),
                            F.col(column_name)
                        ).otherwise(substitute.cast(data_type)))
                    .try_cast(data_type),
                )
            elif mode == "disallow":
                input_df = input_df.withColumn(
                    column_name,
                    F.when(F.col(column_name).isNull(), F.lit(None))
                    .otherwise(F.when(F.col(column_name).isin(elements), substitute).otherwise(F.col(column_name)))
                    .try_cast(data_type),
                )
            elif mode == "NOT_DEFINED":
                raise RuntimeError(
                    f"Please provide a `mode` attribute! Possible values are: ['allow', 'disallow'].\n"
                    f"column_name: {column_name} and cleaning_definition: {cleaning_definition}"
                )
            else:
                raise ValueError(
                    f"Only the following modes are supported by EnumCleaner: ['allow', 'disallow'].\n"
                    f"column_name: {column_name} and cleaning_definition: {cleaning_definition}"
                )

        if self.column_to_log_cleansed_values:
            input_df = self._log_cleansed_values(input_df, column_names_to_clean, temporary_column_names)

        return input_df
